#!/usr/bin/python
"""
Diagnose RFI and RFI removal strategies by plotting either maximum amplitude per channel or fraction of flagged data per channel.
"""
import numpy as np
import aipy as a
from matplotlib import pylab as p
import os,sys,optparse,math

#Set up the options, make sure to define them well.

o = optparse.OptionParser()
o.set_usage('rfi_diagnostic.py [options] *.uvcb')
o.set_description(__doc__)

a.scripting.add_standard_options(o,ant=True,pol=True,chan=False,dec=False,cmap=False,max=False,drng=False)
o.add_option('-m','--max_hold',action='store_true',dest='max_hold',help='Display the maximum amplitude in each frequency bin.')
o.add_option('-d','--dwell_time',action='store_true',dest='dwell',help='Display the fraction of time lost in each frequency bin.')
o.add_option('-x','--x_axis',dest='x_axis',default='freq',help='Indicates the units of the x-axis. freq [default] = frequency (MHz), z = redshift,chan = channel number')
o.add_option('-l','--log',dest='log',action='store_true',help='Display max-hold values in a logarithmic scale. This is recommended.')
o.add_option('-s','--str',dest='strategy',default='normal',help='Chooses the RFI removal strategy. simple = the strategy of xrfi_simple.py. normal [default] = the strategy of xrfi.py. naive = another strategy of xrfi_simple.py, in which those data a certain number of standard deviations above the median are flagged. none = RFI removal has already been performed. Only count flagged data. This is moved to default if filename ends in r or R.')
o.add_option('--best',dest='best',action='store_true',help='Summarizes all baselines. Dispays the median, lowest, and highest average dwell-times (or max-holds) for all selected baselines.')
o.add_option('-g','--gom',dest='gom',default=None,help='Flag the gainometer channels, because we dont care about it.')
o.add_option('--sum',dest='sum',action='store_true',help='Average over all baselines, plot one line.')

opts, args = o.parse_args(sys.argv[1:])

#Run through the uv files and plot what you want to...

plot_y = {}
infected = {}
total = {}
hi_bl,med_bl,lo_bl = '','',''
freqs = []

for uvfile in args:
	print 'Reading',uvfile
	uv = a.miriad.UV(uvfile)
	a.scripting.uv_selector(uv,opts.ant,opts.pol)	
	if uvfile == args[0]:
		freqs = 1000.*np.linspace(uv['sfreq'],uv['sfreq']+uv['nchan']*uv['sdf'],uv['nchan']) #Units are sensible Megahertz.
	if uvfile[-1] == 'r' or uvfile[-1] == 'R': opts.strategy = 'none'
	data,mask = {},{}
	#This is the max-hold part of the process.
	for (uvw,t,(i,j)),d,f in uv.all(raw=True):
		bl = '%d,%d' % (i,j)
		if i==opts.gom or j==opts.gom: continue #Don't look at gainometer channels.
		f = np.where(np.abs(d) == 0,1,f)
		if opts.max_hold:
			if opts.log: d = np.log10(d)
			if not plot_y.has_key(bl): plot_y[bl] = np.zeros(uv['nchan'])
			for I in range(uv['nchan']):
				if d[I] >= plot_y[bl][I]: plot_y[bl][I]=d[I]
		#This is the part that figures out fractional time spent in RFI-mode. Use xrfi filter (derivatives, not polynomial fit) and just compute (flagged_samples/total_samples) for each channel.
		if opts.dwell:
			#Following xrfi.py, record all datums... make a mask later.
			if not data.has_key(bl):
				data[bl]={}
				mask[bl] = {}
			#create data array, mask.
			data[bl][t] = d
			mask[bl][t] = f	
		if not opts.max_hold and not opts.dwell:
			print 'Please indicate whether you want max-hold (-m) or dwell-time (-d) plots'
			sys.exit(0)
	
	if opts.dwell:
		for bl in data: #This is a blatant rip-off of xrfi_simple.py, but that's kind of what we want...	
			data_times = data[bl].keys()
			data_times.sort()	
			rfi_counter = np.array([mask[bl][t] for t in data_times]) 
			
			if opts.strategy == 'simple':
				d = np.array([data[bl][t] for t in data_times])
				#find sigma for frequency.
				ddf = d[:,1:-1]-0.5*(d[:,:-2]+d[:,2:])
				ddf2 = np.abs(ddf)**2
				sigf = np.sqrt(np.median(ddf2,axis=1))
				sigf.shape = (sigf.size,1)
				rfi_counter[:,0] += 1;rfi_counter[:,-1] += 1
				rfi_counter[:,1:-1] += np.where(ddf2/sigf**2 > 4.**2,1,0)
				#same thing, but for time.
				ddt = d[1:-1,:]-0.5*(d[:-2,:]+d[2:,:])
				ddt2 = np.abs(ddt)**2
				sigt = np.sqrt(np.median(ddt2,axis=1))
				sigt.shape = (sigt.size,1)
				rfi_counter[0,:] += 1;rfi_counter[-1,:] += 1
				rfi_counter[1:-1,:] += np.where(ddt2/sigt**2 > 4.**2,1,0)			
	
				rfi_counter = np.where(rfi_counter > 0,1,0)
	
			if opts.strategy == 'naive':
				ad = np.abs(d)
				med = np.median(ad)
				sig = np.sqrt(np.median(np.abs(ad-med)**2))
				rfi_counter += np.where(ad > med +2.*sig,1,0)
				
				rfi_counter = np.where(rfi_counter > 0,1,0) #Don't want to count twice!!!
			
			if opts.strategy == 'normal':
				d = np.ma.array([data[bl][t] for t in data_times],mask=[mask[bl][t] for t in data_times])
				hi,low = a.rfi.gen_rfi_thresh(d)
				m = np.where(np.abs(d) > hi,1,0)
				rfi_counter += m
				ch_cnt = np.array([mask[bl][t] for t in data_times]).sum(axis=0)
				ch_msk = np.where(ch_cnt > ch_cnt.max()*0.33,1,0)
				for I,t in enumerate(rfi_counter):
					if rfi_counter[I].sum() > rfi_counter[I].size*0.33: rfi_counter[I] += np.ones(uv['nchan'])
					else: rfi_counter[I] += ch_msk		
				if i == j:
					bad_ints = a.rfi.flag_by_int(d)
					for I in np.where(bad_ints)[0]:
						rfi_counter[I] += np.ones(uv['nchan'])			
				
				rfi_counter = np.where(rfi_counter > 0,1,0)
			
			if opts.strategy == 'none':
				pass

			#Now, to count everything.
			if not infected.has_key(bl):
				infected[bl] = np.zeros(uv['nchan'])
				total[bl] = np.zeros(uv['nchan'])
			for ct in range(len(rfi_counter)):
				infected[bl] += rfi_counter[ct]
				total[bl] += np.ones(uv['nchan'])	

	del(uv)	

if opts.dwell:
	for bl in infected:
		plot_y[bl] = infected[bl] / total[bl]
		if opts.log: plot_y[bl] = np.log10(plot_y[bl])

#Housekeeping...

#Do some conversion of the frequency axis.
if opts.x_axis == 'freq': 
	plot_x = freqs
	xlabel = 'Frequency (MHz)'
if opts.x_axis == 'z': 
	plot_x = (1427.1/freqs)-1.
	xlabel = 'Redshift of EoR'
if opts.x_axis == 'chan':
	plot_x = np.arange(nchan)
	xlabel = 'Channel Number'

bls = plot_y.keys() #This keeps track of all the baselines.
if len(bls) == 0:
        print 'No data to plot'
        sys.exit(0)

def sort_func(a,b): #This guy will make sure that your baselines always show up in order.
	ai,aj = map(int,a.split(','))
	bi,bj = map(int,b.split(','))
	if bi > ai or (bi == ai and bj > aj): return -1
	return 1
bls.sort(cmp=sort_func)

m2 = int(math.sqrt(len(bls))) #These are to make the plots pretty -- to figure where to place subplots.
m1 = int(math.ceil(float(len(bls))) / m2)

if opts.max_hold and not opts.log: ylabel = 'Max Amplitude (counts)'
if opts.max_hold and opts.log: ylabel = 'Max Amplitude (log10(counts))'
if opts.dwell and not opts.log: ylabel = 'Fractional Dwell Time'
if opts.dwell and opts.log: ylabel = 'Fractional Dwell Time (log)'

#scan through everything if opts.best is chosen...
if opts.best:
	#recover the bl that corresponds the the median, hi, and lo of average dwell_time or max_hold.
	#First, i have to take the averages.
	averages = {}
	hi_bl,med_bl,lo_bl = '','',''
	for bl in plot_y: averages[bl] = np.average(plot_y[bl])
	#I have to do this, because if the number of baselines is even, the median is not contained within averages[]
	values = averages.values()
	values.sort()
	nmed = int(len(averages)/2)
	hi,med,lo = np.max(values),values[nmed],np.min(values)
	for count,bl in enumerate(averages):
		if averages[bl] == hi: hi_bl = bl
		if averages[bl] == med: med_bl = bl
		if averages[bl] == lo: lo_bl = bl
	#Now, plot the three on the same plot...
	p.plot(plot_x,plot_y[lo_bl],'g-',label='Best='+lo_bl)
	p.plot(plot_x,plot_y[med_bl],'b-',label='Median='+med_bl)
	p.plot(plot_x,plot_y[hi_bl],'r-',label='Worst='+hi_bl)
	p.ylabel(ylabel),p.xlabel(xlabel),p.legend()
	p.show()
	sys.exit(0)

if opts.sum:
	av = np.zeros(len(plot_x))
	num = float(len(plot_y))
	for bl in plot_y:
		av += plot_y[bl]
	av /= num
	p.plot(plot_x,av)
	p.xlabel(xlabel);p.ylabel('Average'+ylabel)
	p.show()
	sys.exit(0)

#Now to make the plots.
for counts,bl in enumerate(bls):
	p.subplot(m2,m1,counts+1)
	p.xlim((plot_x[0],plot_x[-1]))
	p.plot(plot_x,plot_y[bl])
	p.ylabel(ylabel);p.xlabel(xlabel);p.title(bl)
p.show()



#!/usr/bin/env python
"""
A script for fitting parameters of a measurement equation given 
starting parameters in a cal file and a list of sources.  The fitter used
here is a steepest-decent filter and does not make use of priors.
"""

import aipy as a, numpy as np, sys, os, optparse

o = optparse.OptionParser()
o.set_usage('fitmdl.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, ant=True, pol=True, chan=True,
    cal=True, src=True, dec=True, prms=True)
o.add_option('-S', '--shared_prms', dest='shprms',
    help='Parameter listing w/ same syntax as "-P/--prms" except that all objects listed for a parameter will share an instance of that parameter.')
o.add_option('--snap', dest='snap', action='store_true',
    help='Snapshot mode.  Fits parameters separately for each integration.')
o.add_option('-q', '--quiet', dest='quiet', action='store_true',
    help='Be less verbose.')
o.add_option('--maxiter', dest='maxiter', type='float', default=-1,
    help='Maximum # of iterations to run.  Default is infinite.')
o.add_option('--xtol', dest='xtol', type='float', default=1e-10,
    help='Fractional change sought in it parameters before convergence.  Default 1e-10.')
o.add_option('--ftol', dest='ftol', type='float', default=1e-10,
    help='Fractional tolerance sought in score before convergence.  Default 1e-10.')
o.add_option('--remem', dest='remember', action='store_true',
    help='Remember values from last fit when fitting in snapshot mode.')
o.add_option('--baseport', dest='baseport', type='int', default=53000,
    help="Base port # to use for tx/rx.  Each daemon adds it's daemon id to this to determine the actual port used for TCP transactions.")
o.add_option('--daemon', dest='daemon', type='int', 
    help='Operate in daemon mode, opening a TCP Server to handle requests on the specified increment to the base port.')
o.add_option('--master', dest='master', 
    help='Operate in master mode, employing daemon-mode servers to do the work and collecting the results.  Should be a comma delimited list of host:daemonid pairs to contact.  Daemon ID will be added to baseport to determine actual port used for TCP transactions.')
o.add_option('--sim_autos', dest='sim_autos', action='store_true',
    help='Use auto-correlations in fitting.  Default is to use only cross-correlations.')
o.add_option('--minuv',dest='minuv',default=0.,type='float',help='Minimum baseline length (in wavelengths at 150 MHz) to consider.')

opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
#opts.ant += ',cross'
a.scripting.uv_selector(uv, opts.ant, opts.pol)
aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
aa.select_chans(chans)
srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
(uvw,t,(i,j)),d = uv.read()
aa.set_jultime(t)
cat.compute(aa)
del(uv)
if opts.maxiter < 0: opts.maxiter = np.Inf

# Figure out parameters to fit
prms, prm_dict, shkeys = {}, {}, []
# Handle shared parameters
if opts.shprms:
    shprms = map(a.scripting.parse_prms, opts.shprms.split(','))
    for s in shprms:
        keys = s.keys(); keys.sort()
        k = keys[0]
        # Only enter one instance of shared parameters (indexed under first key)
        if prms.has_key(k): prms[k].update(s[k])
        else: prms.update({k:s[k]})
        # Add an entry to shkeys for propagating variables to other objects
        shkeys.append((keys, s[k].keys()))
# Handle normal parameters
if opts.prms:
    pd = a.scripting.parse_prms(opts.prms)
    for k in pd:
        if prms.has_key(k): prms[k].update(pd[k])
        else: prms[k] = pd[k]
for prm in prms: prm_dict[prm] = prms[prm].keys()
start_prms = aa.get_params(prm_dict)
start_prms.update(cat.get_params(prm_dict))
for obj in start_prms:
    for prm in start_prms[obj]:
        if prms[obj][prm][0] != None:
            start_prms[obj][prm] = prms[obj][prm][0]
        
prm_list, key_list = a.fit.flatten_prms(start_prms)

first_fit = None    # Used to normalize fit values to the starting fit
mfq = cat.get('mfreq')
dbuf = None

def uvlen(A): return np.sqrt(np.dot(A,A))

# The function to be optimized
def fit_func(prms, filelist, decimate, decphs):
    global first_fit, dbuf
    if first_fit == 0: return 0
    prms = a.fit.reconstruct_prms(prms, key_list)
    # Propagate shared params
    for (skey,sprm) in shkeys:
        k = skey[0]
        for k2 in skey[1:]:
            if not prms.has_key(k2): prms[k2] = {}
            for sp in sprm:
                prms[k2][sp] = prms[k][sp]
    if not opts.quiet:
        a.fit.print_params(prms)
        print prms
    aa.set_params(prms)
    cat.set_params(prms)
    a1,a2,th = cat.get('srcshape')
    score,nsamples = 0.,0.
    # Cache data from file to avoid hitting disk excessively
    if dbuf is None:
        if not opts.quiet: print 'Caching data...'
        dbuf = {}
        for uvfile in filelist:
            sys.stdout.write('.') ; sys.stdout.flush()
            uv = a.miriad.UV(uvfile)
            a.scripting.uv_selector(uv, opts.ant, opts.pol)
            uv.select('decimate', decimate, decphs)
            for (uvw,t,(i,j)),d,f in uv.all(raw=True):
                bl = a.miriad.ij2bl(i,j)
                if not dbuf.has_key(t): dbuf[t] = {}
                if not dbuf[t].has_key(bl): dbuf[t][bl] = {}
                if not opts.sim_autos and i == j: continue
                if uvlen(aa.get_baseline(i,j))*0.15 < opts.minuv: continue
                d = d.take(chans)
                f = f.take(chans)
                pol = a.miriad.pol2str[uv['pol']]
                dbuf[t][bl][pol] = (d, f, np.where(f, 0, np.abs(d)**2).sum())
        if not opts.quiet:
            samp, vsamp, wgts = 0, 0, 0.
            for t in dbuf:
              for bl in dbuf[t]:
                for pol in dbuf[t][bl]:
                    samp += len(dbuf[t][bl][pol][1])
                    vsamp += np.logical_not(dbuf[t][bl][pol][1]).astype(np.int).sum()
                    wgts += dbuf[t][bl][pol][2]
            print 'Cache summary:'
            print '   %d samples' % samp
            print '   %d valid' % vsamp
            print '   %f sum weights' %  wgts
            sys.stdout.flush()
    # Process data from cache
    for t in dbuf:
        aa.set_jultime(t)
        cat.compute(aa)
        eqs = cat.get_crds('eq', ncrd=3)
        flx = cat.get_jys()
        dra,ddec = cat.get('ionref')
        aa.sim_cache(eqs, flx, mfreqs=mfq, 
            ionrefs=(dra,ddec), srcshapes=(a1,a2,th))
        for bl in dbuf[t]:
            i,j = a.miriad.bl2ij(bl)
            for pol in dbuf[t][bl]:
                d,f,nsamp = dbuf[t][bl][pol]
                aa.set_active_pol(pol)
                sim_d = aa.sim(i, j)
                difsq = np.abs(d - sim_d)**2
                difsq = np.where(f, 0, difsq)
                score += difsq.sum()
                nsamples += nsamp
    if opts.daemon: return score, nsamples
    if nsamples == 0:
        first_fit = 0.
        return 0.
    score = np.sqrt(score / nsamples)
    if first_fit is None: first_fit = score
    if not opts.quiet:
        print
        print 'Score:', score, 
        print '(%2.2f%% of %f)' % (100 * score / first_fit, first_fit)
        print '-' * 70
    return score / first_fit

# Call the optimizer
if opts.daemon:
    import SocketServer, struct
    class TCPServer(SocketServer.TCPServer):
        allow_reuse_address = True
    class FitHandler(SocketServer.BaseRequestHandler):
        def setup(self): print self.client_address, 'connected'
        def finish(self): print self.client_address, 'disconnected'
        def handle(self):
            data = self.request.recv(struct.calcsize('d')*len(prm_list))
            data = struct.unpack('<%dd' % (len(prm_list)), data)
            score, nsamples = fit_func(data, args, opts.decimate, opts.decphs)
            print 'Returning score=%f, nsamples=%f' % (score, nsamples)
            rv = struct.pack('<dd', score, nsamples)
            self.request.send(rv)
            sys.stdout.flush()
    s = TCPServer(('', opts.baseport + opts.daemon), FitHandler)
    print 'Starting daemon on TCP port %d' % (opts.baseport + opts.daemon)
    sys.stdout.flush()
    s.serve_forever()
elif opts.master:
    import socket, struct
    def parsehostport(hostport):
        host, port = hostport.split(':')
        return (host, opts.baseport + int(port))
    hostports = [parsehostport(w) for w in opts.master.split(',')]
    def fit_func(prms):
        global first_fit
        if first_fit == 0: return 0
        pdict = a.fit.reconstruct_prms(prms, key_list)
        if not opts.quiet: a.fit.print_params(pdict)
        data = struct.pack('<%dd' % (len(prms)), *prms)
        socks = []
        for hostport in hostports:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect(hostport)
            sock.send(data)
            socks.append(sock)
        score, nsamples = 0, 0
        for sock in socks:
            data = sock.recv(1024)
            scr, nsp = struct.unpack('<dd', data)
            score += scr; nsamples += nsp
            if not opts.quiet: print score, nsamples
        score = np.sqrt(score / nsamples)
        if nsamples == 0:
            first_fit = 0.
            return 0.
        if first_fit is None: first_fit = score
        if not opts.quiet:
            print
            print 'Score:', score, 
            print '(%2.2f%% of %f)' % (100 * score / first_fit, first_fit)
            print '-' * 70
        return score / first_fit
    print 'Starting in master mode...'
    rv = a.optimize.fmin(
        fit_func, prm_list,
        #args=(args, opts.decimate, opts.decphs),
        full_output=1, disp=0,
        maxfun=opts.maxiter, maxiter=np.Inf, 
        ftol=opts.ftol, xtol=opts.xtol
    )
    prms,score = rv[:2]
    prms = a.fit.reconstruct_prms(prms, key_list)
    print
    a.fit.print_params(prms)
    print 'Score:', score * first_fit, 
    print '(%2.2f%% of %f)' % (100 * score, first_fit)
    print '------------------------------------------------------------'
            
    

elif not opts.snap:
    rv = a.optimize.fmin(
        fit_func, prm_list,
        args=(args, opts.decimate, opts.decphs),
        full_output=1, disp=0,
        maxfun=opts.maxiter, maxiter=np.Inf, 
        ftol=opts.ftol, xtol=opts.xtol
    )
    prms,score = rv[:2]
    prms = a.fit.reconstruct_prms(prms, key_list)
    print
    a.fit.print_params(prms)
    print 'Score:', score * first_fit, 
    print '(%2.2f%% of %f)' % (100 * score, first_fit)
    print '------------------------------------------------------------'
else:
    for uvfile in args:
        # Figure out what times are in the file
        uv = a.miriad.UV(uvfile)
        a.scripting.uv_selector(uv, opts.ant, opts.pol)
        uv.select('decimate', opts.decimate, opts.decphs)
        times = [0]
        for (uvw,t,(i,j)),d,f in uv.all(raw=True):
            if times[-1] != t: times.append(t)
        times = times[1:]
        del(uv)
        decimate = len(times) * opts.decimate
        # Fit each time separately
        for cnt, t in enumerate(times):
            print 'Time:', t
            print 'Iter: %d / %d' % (cnt+1, len(times))
            first_fit,dbuf = None,None
            rv = a.optimize.fmin(
                fit_func, prm_list,
                args=([uvfile], decimate, opts.decimate*cnt + opts.decphs),
                full_output=1, disp=0,
                maxfun=opts.maxiter, maxiter=np.Inf, 
                ftol=opts.ftol, xtol=opts.xtol
            )
            prms,score = rv[:2]
            prms = a.fit.reconstruct_prms(prms, key_list)
            print
            print prms
            a.fit.print_params(prms)
            print 'Score:', score * first_fit, 
            print '(%2.2f%% of %f)' % (100 * score, first_fit)
            print '------------------------------------------------------------'
            if opts.remember: prm_list, key_list = a.fit.flatten_prms(prms)



"""
Optimization Tools
==================

 A collection of general-purpose optimization routines.

   fmin        --  Nelder-Mead Simplex algorithm
                     (uses only function calls)
   fmin_powell --  Powell's (modified) level set method (uses only
                     function calls)
   fmin_cg     --  Non-linear (Polak-Ribiere) conjugate gradient algorithm
                     (can use function and gradient).
   fmin_bfgs   --  Quasi-Newton method (Broydon-Fletcher-Goldfarb-Shanno); 
                     (can use function and gradient)
   fmin_ncg    --  Line-search Newton Conjugate Gradient (can use
                     function, gradient and Hessian).
   leastsq     --  Minimize the sum of squares of M equations in
                     N unknowns given a starting estimate.

  Global Optimizers

   anneal      --  Simulated Annealing
   brute       --  Brute force searching optimizer


  Scalar function minimizers

   fminbound   --  Bounded minimization of a scalar function.
   brent       --  1-D function minimization using Brent method.
   golden      --  1-D function minimization using Golden Section method
   bracket     --  Bracket a minimum (given two starting points)


 Also a collection of general-purpose root-finding routines.

   fsolve      --  Non-linear multi-variable equation solver.


  Scalar function solvers

   brentq      --  quadratic interpolation Brent method
   brenth      --  Brent method (modified by Harris with hyperbolic
                     extrapolation)
   ridder      --  Ridder's method
   bisect      --  Bisection method
   newton      --  Secant method or Newton's method

   fixed_point --  Single-variable fixed-point solver.

 A collection of general-purpose nonlinear multidimensional solvers.

   broyden1            --  Broyden's first method - is a quasi-Newton-Raphson
                           method for updating an approximate Jacobian and then
                           inverting it
   broyden2            --  Broyden's second method - the same as broyden1, but
                           updates the inverse Jacobian directly
   broyden3            --  Broyden's second method - the same as broyden2, but
                           instead of directly computing the inverse Jacobian,
                           it remembers how to construct it using vectors, and
                           when computing inv(J)*F, it uses those vectors to
                           compute this product, thus avoding the expensive NxN
                           matrix multiplication.  
   broyden_generalized --  Generalized Broyden's method, the same as broyden2,
                           but instead of approximating the full NxN Jacobian,
                           it construct it at every iteration in a way that
                           avoids the NxN matrix multiplication.  This is not
                           as precise as broyden3.
   anderson            --  extended Anderson method, the same as the
                           broyden_generalized, but added w_0^2*I to before
                           taking inversion to improve the stability
   anderson2           --  the Anderson method, the same as anderson, but
                           formulated differently

 Utility Functions

   line_search --  Return a step that satisfies the strong Wolfe conditions.
   check_grad  --  Check the supplied derivative using finite difference
                     techniques.
"""




"""
Module for mapping and modeling the entire sky.
"""

import numpy as np, healpix, coord, random,img
try:
    from astropy.io import fits as pyfits
except ImportError:
    import pyfits

# Set a fixed random seed to make scrambling deterministic
random.seed(1)

deg2rad = np.pi / 180.
rad2deg = 180. / np.pi

def pack_sphere(N):
    """Clever formula for putting N points nearly equally spaced on 
    the sphere.  Return xyz coordinates of each point in order from S to N."""
    dz = 2. / N
    z = np.arange(-1+dz/2,1, dz)
    r = np.sqrt(1-z**2)
    dL = np.pi * (3 - np.sqrt(5))
    long = np.arange(0, dL * N, dL)
    return np.array([r*np.cos(long), r*np.sin(long), z])

def _bit_reverse(N, nbits=None):
    """Numerically bit-reverse the number N assuming the specified number
    of bits.  In not specified, will infer the least number of bits that
    can represent N."""
    if nbits is None: nbits = int(np.floor(np.log2(N))) + 1
    ans = 0
    for bit in range(nbits):
        ans += np.bitwise_and(N, 2**bit) * 2**(nbits-2*bit-1)
    return ans

def _bit_reverse_order(N):
    """Generate a list of indices from 0 to N-1 in bit reversed order (or
    the nearest approximation if N is not a power of 2)."""
    nbits = int(np.floor(np.log2(N))) + 1
    indices = _bit_reverse(np.arange(2**nbits), nbits=nbits)
    return indices.compress(indices < N)

def _local_shuffle(L, width=2):
    """Shuffle elements of L without moving them too much by choosing chunks
    of the specified width and only scrambling elements within those chunks."""
    for i in range(int(np.ceil(len(L) / float(width)))):
        chunk = L[width*i:width*(i+1)]
        random.shuffle(chunk)
        L[width*i:width*(i+1)] = chunk

def facet_centers(N, ncrd=2):
    """Return the coordinates of N points equally spaced around the sphere.
    Will return xyz or ra,dec depending on ncrd.  Shuffles the order of the
    pointing centers wrt pack_sphere so that widely spaced points are done 
    first, and then the gaps between them, and then the gaps between those.."""
    assert(ncrd == 2 or ncrd == 3)
    ind1 = np.arange(N); _local_shuffle(ind1)
    ind2 = _bit_reverse_order(N)
    ind = ind1.take(ind2)
    pnts = pack_sphere(N)
    pnts = pnts.take(ind, axis=1)
    if ncrd == 3: return pnts
    else: return coord.eq2radec(pnts)

class Map(object):
    def __init__(self, *args, **kwargs):
        fromfits = kwargs.pop('fromfits', None)
        nindices = kwargs.pop('nindices', None)
        self.args,self.kwargs = args,kwargs
        self.map = healpix.HealpixMap(*args, **kwargs)
        self.wgt = healpix.HealpixMap(*args, **kwargs)
        self.ind = []
        if not fromfits is None: self.from_fits(fromfits)
        self.set_nindices(nindices)
    def set_interpol(self, onoff):
        self.map.set_interpol(onoff)
        self.wgt.set_interpol(onoff)
        for i in self.ind: i.set_interpol(onoff)
    def set_nindices(self, nindices):
        if nindices is None: return
        self.ind = self.ind[:nindices]
        for i in range(nindices - len(self.ind)):
            self.ind.append(healpix.HealpixMap(*self.args, **self.kwargs))
    def get(self, crds):
        fluxes = self.map[crds]
        wgts = self.wgt[crds]
        inds = [i[crds] for i in self.ind]
        return (wgts, fluxes, inds)
    def __getitem__(self, crds):
        """Return the average map/index values at the specified coordinates."""
        w = self.wgt[crds]
        w = np.where(w > 0, w, 1)
        fluxes = self.map[crds] / w
        ind = [i[crds] / w for i in self.ind]
        if len(ind) == 0: return fluxes
        return (fluxes, ind)
    def add(self, crds, wgts, fluxes, inds=[]):
        self.wgt[crds] += wgts
        self.map[crds] += fluxes * wgts
        for n,i in enumerate(inds): self.ind[n][crds] += i * wgts
    def put(self, crds, wgts, fluxes, inds=[]):
        self.wgt[crds] = wgts
        self.map[crds] = fluxes * wgts
        for n,i in enumerate(inds): self.ind[n][crds] = i * wgts
    def reset_wgt(self, wgt=1):
        w = np.where(self.wgt.map > 0, self.wgt.map, 1)
        self.map.map /= w
        for i in self.ind: i.map /= w
        self.wgt.map = np.where(self.wgt.map > 0, wgt, 0)
    def from_map(self, map):
        """Initialize this Map with data from another."""
        self.map.from_hpm(map.map)
        self.wgt.from_hpm(map.wgt)
        for i in range(min(len(self.ind), len(map.ind))): 
            self.ind[i].from_hpm(map.ind[i])
    def __getattr__(self, attr):
        try: object.__getatr__(self, attr)
        except(AttributeError): return self.map.__getattribute__(attr)
    def from_fits(self, filename, hdunum=1):
        self.map.from_fits(filename, hdunum=hdunum, colnum=0)
        self.args = ()
        self.kwargs = {'nside':self.nside(), 'scheme':self.scheme()}
        try:
            self.wgt.from_fits(filename, hdunum=hdunum, colnum=1)
        except(IndexError):
            self.wgt = healpix.HealpixMap(*self.args, **self.kwargs)
            self.wgt.set_map(np.ones_like(self.wgt.map))
        # Figure out how many cols there are (is there a better way?)
        hdu = pyfits.open(filename)[hdunum]
        nind = 0
        while True:
            try: i = hdu.data.field(2 + nind)
            except(IndexError): break
            nind += 1
        del(hdu)
        # Make a spectral index HPM for each additional col in fits file
        for i in range(nind):
            h = healpix.HealpixMap(*self.args, **self.kwargs)
            h.from_fits(filename, hdunum=hdunum, colnum=2+i)
            self.ind.append(h)
    def to_fits(self, filename, format=None, clobber=False,history=''):
        if format is None:
            format = healpix.default_fits_format_codes[self.get_dtype().type]
        hdu0 = pyfits.PrimaryHDU()
        col0 = pyfits.Column(name='signal', format=format, array=self.map.map)
        col1 = pyfits.Column(name='weights', format=format, array=self.wgt.map)
        col_inds = [pyfits.Column(name='sp_index%d'%n,format=format,array=i.map)
            for n,i in enumerate(self.ind)]
        cols = pyfits.ColDefs([col0, col1] + col_inds)
        tbhdu = pyfits.new_table(cols)
        self.map._set_fits_header(tbhdu.header)
        hdulist = pyfits.HDUList([hdu0, tbhdu])
        if history!='':
            history = [h.strip() for h in history.split("\n")]
            for line in history:
                if len(line)>1:
                    if line.startswith('#'):
                        for subline in img.word_wrap(line,80,0,0,'').split("\n"):
                            hdulist[0].header.add_history(subline)
                    else:
                        for subline in img.word_wrap(line,70,5,10,'#').split("\n"):
                            hdulist[0].header.add_history(subline)       
        hdulist.writeto(filename, clobber=clobber)



#! /usr/bin/env python
"""
Models visibilities for various catalog sources and creates a new Miriad UV
file containing either the simulated data, or the residual when the model
is removed from measured data.

Author: Aaron Parsons
"""
import numpy as np, aipy as a, optparse, os, sys, ephem

o = optparse.OptionParser()
o.set_usage('mdlvis.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, ant=True, cal=True, src=True)
o.add_option('-m','--mode', dest='mode', default='sim',
    help='Operation mode.  Can be "sim" (output simulated data), "sub" (subtract from input data), or "add" (add to input data).  Default is "sim"')
o.add_option('-f', '--flag', dest='flag', action='store_true',
    help='If outputting a simulated data set, mimic the data flagging of the original dataset.')
#o.add_option('-m', '--map', dest='map',
#    help='The Healpix map to use for simulation input.')
#o.add_option('--iepoch', dest='iepoch', default=ephem.J2000, 
#    help='The epoch of coordinates in the map. Default J2000.')
#o.add_option('--freq', dest='freq', default=.150, type='float',
#    help='Frequency of flux data in map.')
o.add_option('-n', '--noiselev', dest='noiselev', default=0., type='float',
    help='RMS amplitude of noise (Jy) added to each UV sample of simulation.')
o.add_option('--nchan', dest='nchan', default=256, type='int',
    help='Number of channels in simulated data if no input data to mimic.  Default is 256')
o.add_option('--sfreq', dest='sfreq', default=.075, type='float',
    help='Start frequency (GHz) in simulated data if no input data to mimic.  Default is 0.075')
o.add_option('--sdf', dest='sdf', default=.150/256, type='float',
    help='Channel spacing (GHz) in simulated data if no input data to mimic.  Default is .150/256')
o.add_option('--inttime', dest='inttime', default=10, type='float',
    help='Integration time (s) in simulated data if no input data to mimic.  Default is 10')
o.add_option('--startjd', dest='startjd', default=2454600., type='float',
    help='Julian Date to start observation if no input data to mimic.  Default is 2454600')
o.add_option('--endjd', dest='endjd', default=2454601., type='float',
    help='Julian Date to end observation if no input data to mimic.  Default is 2454601')
o.add_option('--pol', dest='pol', 
    help='Polarizations to simulate (xx,yy,xy,yx) if starting file from scratch.')
opts, args = o.parse_args(sys.argv[1:])

assert(len(args) > 0 or (opts.mode == 'sim' and not opts.flag and not (opts.pol is None)))
# Parse command-line options
if len(args) > 0:
    uv = a.miriad.UV(args[0])
    aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
    p,d,f = uv.read(raw=True)
    no_flags = np.zeros_like(f)
    del(uv)
else:
    aa = a.cal.get_aa(opts.cal, opts.sdf, opts.sfreq, opts.nchan)
    no_data = np.zeros(opts.nchan, dtype=np.complex64)
    no_flags = np.zeros(opts.nchan, dtype=np.int32)

# Generate a model of the sky with point sources and a pixel map

#mfq,a1s,a2s,ths = [], [], [], []
#dras,ddecs = [], []
# Initialize point sources
#if not opts.src is None:
srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
mfq = cat.get('mfreq')
a1s,a2s,ths = cat.get('srcshape')
#a1s.append(a1); a2s.append(a2); ths.append(th)
dras, ddecs = cat.get('ionref')
#dras.append(dra); ddecs.append(ddec)

## Initialize pixel map
#if not opts.map is None:
#    h = a.map.Map(fromfits=opts.map)
#    px = np.arange(h.npix())
#    try: mflx, i_poly = h[px]
#    except(ValueError): mflx = h[px]
#    px = np.compress(mflx > 0, px)
#    try: mflx, i_poly = h[px]
#    except(ValueError):
#        mflx = h[px]
#        i_poly = [np.zeros_like(mflx)]
#    mind = i_poly[0]    # Only implementing first index term for now
#    mmfq = opts.freq * np.ones_like(mind)
#    mfq.append(mmfq)
#    x,y,z = h.px2crd(px, ncrd=3)
#    m_eq = np.array((x,y,z))
#    # Should pixels include information for resolving them?
#    a1s.append(np.zeros_like(mmfq))
#    a2s.append(np.zeros_like(mmfq))
#    ths.append(np.zeros_like(mmfq))
#    dras.append(np.zeros_like(mmfq))
#    ddecs.append(np.zeros_like(mmfq))
#mfq = np.concatenate(mfq)
#a1s = np.concatenate(a1s)
#a2s = np.concatenate(a2s)
#ths = np.concatenate(ths)
#dras = np.concatenate(dras)
#ddecs = np.concatenate(ddecs)

# A pipe for applying the model
curtime = None
def mdl(uv, p, d, f):
    global curtime, eqs
    uvw, t, (i,j) = p
    pol = a.miriad.pol2str[uv['pol']]
    if i == j: return p, d, f
    if curtime != t:
        curtime = t
        aa.set_jultime(t)
        #eqs,flx,ind = [],[],[]
        #if not opts.src is None:
        cat.compute(aa)
        eqs = cat.get_crds('eq', ncrd=3)
        flx = cat.get_jys()
        #if not opts.map is None:
        #    m_precess = a.coord.convert_m('eq','eq',
        #        iepoch=opts.iepoch, oepoch=aa.epoch)
        #    eqs.append(np.dot(m_precess, m_eq))
        #    flx.append(mflx); ind.append(mind)
        #eqs = np.concatenate(eqs, axis=-1)
        #flx = np.concatenate(flx)
        #ind = np.concatenate(ind)
        aa.sim_cache(eqs, flx, mfreqs=mfq, 
            ionrefs=(dras,ddecs), srcshapes=(a1s,a2s,ths))
    aa.set_active_pol(pol)
    sd = aa.sim(i, j)
    if opts.mode.startswith('sim'):
        d = sd
        if not opts.flag: f = no_flags
    elif opts.mode.startswith('sub'):
        d -= sd
    elif opts.mode.startswith('add'):
        d += sd
    else:
        raise ValueError('Mode "%s" not supported.' % opts.mode)
    if opts.noiselev != 0:
        # Add on some noise for a more realistic experience
        noise_amp = np.random.random(d.shape) * opts.noiselev
        noise_phs = np.random.random(d.shape) * 2*np.pi * 1j
        noise = noise_amp * np.exp(noise_phs)
        d += noise * aa.passband(i,j)
    return p, np.where(f, 0, d), f

if len(args) > 0:
    # Run mdl on all files
    for filename in args:
        uvofile = filename + 's'
        print filename,'->',uvofile
        if os.path.exists(uvofile):
            print 'File exists: skipping'
            continue
        uvi = a.miriad.UV(filename)
        a.scripting.uv_selector(uvi, opts.ant)
        uvo = a.miriad.UV(uvofile, status='new')
        uvo.init_from_uv(uvi)
        uvo.pipe(uvi, mfunc=mdl, raw=True, append2hist="MDLVIS: " + ' '.join(sys.argv) + '\n')
else:
    # Initialize a new UV file
    pols = opts.pol.split(',')
    uv = a.miriad.UV('new.uv', status='new')
    uv._wrhd('obstype','mixed-auto-cross')
    uv._wrhd('history','MDLVIS: created file.\nMDLVIS: ' + ' '.join(sys.argv) + '\n')
    uv.add_var('telescop','a'); uv['telescop'] = 'AIPY'
    uv.add_var('operator','a'); uv['operator'] = 'AIPY'
    uv.add_var('version' ,'a'); uv['version'] = '0.0.1'
    uv.add_var('epoch'   ,'r'); uv['epoch'] = 2000.
    uv.add_var('source'  ,'a'); uv['source'] = 'zenith'
    uv.add_var('latitud' ,'d'); uv['latitud'] = aa.lat
    uv.add_var('dec'     ,'d'); uv['dec'] = aa.lat
    uv.add_var('obsdec'  ,'d'); uv['obsdec'] = aa.lat
    uv.add_var('longitu' ,'d'); uv['longitu'] = aa.long
    uv.add_var('npol'    ,'i'); uv['npol'] = len(pols)
    uv.add_var('nspect'  ,'i'); uv['nspect'] = 1
    uv.add_var('nants'   ,'i'); uv['nants'] = len(aa)
    uv.add_var('antpos'  ,'d')
    antpos = np.array([ant.pos for ant in aa], dtype=np.double)
    uv['antpos'] = antpos.transpose().flatten()
    uv.add_var('sfreq'   ,'d'); uv['sfreq'] = opts.sfreq
    uv.add_var('freq'    ,'d'); uv['freq'] = opts.sfreq
    uv.add_var('restfreq','d'); uv['restfreq'] = opts.sfreq
    uv.add_var('sdf'     ,'d'); uv['sdf'] = opts.sdf
    uv.add_var('nchan'   ,'i'); uv['nchan'] = opts.nchan
    uv.add_var('nschan'  ,'i'); uv['nschan'] = opts.nchan
    uv.add_var('inttime' ,'r'); uv['inttime'] = float(opts.inttime)
    # These variables just set to dummy values
    uv.add_var('vsource' ,'r'); uv['vsource'] = 0.
    uv.add_var('ischan'  ,'i'); uv['ischan'] = 1
    uv.add_var('tscale'  ,'r'); uv['tscale'] = 0.
    uv.add_var('veldop'  ,'r'); uv['veldop'] = 0.
    # These variables will get updated every spectrum
    uv.add_var('coord'   ,'d')
    uv.add_var('time'    ,'d')
    uv.add_var('lst'     ,'d')
    uv.add_var('ra'      ,'d')
    uv.add_var('obsra'   ,'d')
    uv.add_var('baseline','r')
    uv.add_var('pol'     ,'i')

    # Now start generating data
    times = np.arange(opts.startjd, opts.endjd, opts.inttime/a.const.s_per_day)
    for cnt,t in enumerate(times):
        print 'Timestep %d / %d' % (cnt+1, len(times))
        aa.set_jultime(t)
        uv['lst'] = aa.sidereal_time()
        uv['ra'] = aa.sidereal_time()
        uv['obsra'] = aa.sidereal_time()
        # XXX should really honor opts.ants for which antennas to simulate here.
        for i,ai in enumerate(aa):
            for j,aj in enumerate(aa):
                try:
                    if ai.pol+aj.pol not in pols: continue
                    i,j = ai.num,aj.num
                except(AttributeError): pass
                if j < i: continue
                crd = aa.get_baseline(i,j)
                preamble = (crd, t, (i,j))
                for pol in pols:
                    uv['pol'] = a.miriad.str2pol[pol]
                    preamble,data,flags = mdl(uv, preamble, None, None)
                    if data is None:
                        data = no_data
                        flags = no_flags
                    uv.write(preamble, data, flags)
    del(uv)



#! /usr/bin/env python
"""
Rotate zenith UV data to a particular source.  Can specify 'zen' to phase data
to zenith, or nothing at all to just remove delay/offset phase components.
"""

import aipy as a, numpy as np, sys, os, optparse

o = optparse.OptionParser()
o.set_usage('phs2src.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, cal=True, src=True)
o.add_option('--setphs', dest='setphs', action='store_true',
    help='Instead of rotating phase, assign a phase corresponding to the specified source.')
o.add_option('--rot_uvw',action='store_true',
    help='Rotate the uvw coordinates to the source. Useful for exporting beyond AIPY')
opts,args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
if not opts.src is None:
    if not opts.src.startswith('zen'):
        srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
        src = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs).values()[0]
    else: src = 'z'
else: src = None
del(uv)

# A pipe to use for phasing to a source
curtime = None
def phs(uv, p, d, f):
    global curtime
    uvw, t, (i,j) = p
    pol = a.miriad.pol2str[uv['pol']]
    aa.set_active_pol(pol)
    if curtime != t:
        curtime = t
        aa.set_jultime(t)
        if not src is None and not type(src) == str: src.compute(aa)
    if i == j: return p, d, f
    try:
        if opts.setphs: d = aa.unphs2src(np.abs(d), src, i, j)
        elif src is None: d *= np.exp(-1j*np.pi*aa.get_phs_offset(i,j))
        else: d = aa.phs2src(d, src, i, j)
        if opts.rot_uvw: 
            uvw = aa.gen_uvw(i,j,src=src)
            p = (uvw,t,(i,j))
    except(a.phs.PointingError): d *= 0
    return p, d, f

# Process data
for filename in args:
    if not opts.src is None: uvofile = filename + '.' + opts.src
    else: uvofile = filename + 'P'
    print filename,'->',uvofile
    if os.path.exists(uvofile):
        print 'File exists: skipping'
        continue
    uvi = a.miriad.UV(filename)
    uvo = a.miriad.UV(uvofile, status='new')
    uvo.init_from_uv(uvi)
    uvo.pipe(uvi, mfunc=phs, raw=True)




"""
Module adding simulation support to RadioBodys and AntennaArrays.
Mostly, this means adding gain/amplitude information vs. frequency.
"""

import numpy as np
import ephem
from . import phs
from . import coord
from . import healpix

#  ____           _ _       ____            _       
# |  _ \ __ _  __| (_) ___ | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \|  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|____/ \___/ \__,_|\__, |
#                                             |___/ 

class RadioBody:
    """Class defining flux and spectral index of a celestial source."""
    def __init__(self, jys, index):
        """jys = source strength in Janskies
        mfreq = frequency (in GHz) where strength was measured
        index = index of power-law spectral model of source emission."""
        self._jys = jys
        self.index = index
    def update_jys(self, afreqs):
        """Update fluxes relative to the provided observer.  Must be
        called at each time step before accessing information."""
        self.afreqs = afreqs
        self.jys = self._jys * (afreqs / self.mfreq)**self.index
    def get_jys(self):
        """Return the fluxes vs. freq that should be used for simulation."""
        return self.jys

#  ____           _ _       _____ _              _ ____            _       
# |  _ \ __ _  __| (_) ___ |  ___(_)_  _____  __| | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \| |_  | \ \/ / _ \/ _` |  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) |  _| | |>  <  __/ (_| | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|_|   |_/_/\_\___|\__,_|____/ \___/ \__,_|\__, |
#                                                                    |___/ 

class RadioFixedBody(phs.RadioFixedBody, RadioBody):
    """Class representing a source at fixed RA,DEC.  Adds flux information to
    phs.RadioFixedBody."""
    def __init__(self, ra, dec, name='', epoch=ephem.J2000,
            jys=0., index=-1, mfreq=.150,
            ionref=(0.,0.), srcshape=(0.,0.,0.), **kwargs):
        """ra = source's right ascension (epoch=J2000)
        dec = source's declination (epoch=J2000)
        jys = source strength in Janskies at mfreq)
        mfreq = frequency (in GHz) where source strength was measured
        index = power-law index of source emission vs. freq."""
        phs.RadioFixedBody.__init__(self, ra, dec, mfreq=mfreq, name=name,
            epoch=epoch, ionref=ionref, srcshape=srcshape)
        RadioBody.__init__(self, jys, index)
    def __str__(self):
        outstr = phs.RadioFixedBody.__str__(self)
        try: outstr = outstr +'\t'+ str(self._jys) 
        except(AttributeError): outstr = outstr + '\t'
        try: outstr += '\t' + str(self.index)
        except(AttributeError): outstr += '\t'  
        outstr += '\t'+str(self.mfreq)
        try: outstr = outstr +'\t'+ str(self.jys[0]) 
        except(AttributeError): outstr = outstr + '\t'
        outstr += '\t'+str(self.afreqs[0])
        return outstr
    def compute(self, observer):
        phs.RadioFixedBody.compute(self, observer)
        RadioBody.update_jys(self, observer.get_afreqs())

#  ____           _ _      ____                  _       _ 
# |  _ \ __ _  __| (_) ___/ ___| _ __   ___  ___(_) __ _| |
# | |_) / _` |/ _` | |/ _ \___ \| '_ \ / _ \/ __| |/ _` | |
# |  _ < (_| | (_| | | (_) |__) | |_) |  __/ (__| | (_| | |
# |_| \_\__,_|\__,_|_|\___/____/| .__/ \___|\___|_|\__,_|_|
#                               |_|                        

class RadioSpecial(phs.RadioSpecial, RadioBody):
    """Class representing moving sources (Sun,Moon,planets).  Adds flux
    information to phs.RadioSpecial."""
    def __init__(self, name, jys=0., index=-1., mfreq=.150,
            ionref=(0.,0.), srcshape=(0.,0.,0.), **kwargs):
        """jys = source strength in Janskies at mfreq)
        mfreq = frequency (in GHz) where source strength was measured
        index = power-law index of source emission vs. freq."""
        phs.RadioSpecial.__init__(self, name, mfreq=mfreq,
            ionref=ionref, srcshape=srcshape)
        RadioBody.__init__(self, jys, index)
    def compute(self, observer):
        phs.RadioSpecial.compute(self, observer)
        RadioBody.update_jys(self, observer.get_afreqs())

#  ____            ____      _        _             
# / ___| _ __ ___ / ___|__ _| |_ __ _| | ___   __ _ 
# \___ \| '__/ __| |   / _` | __/ _` | |/ _ \ / _` |
#  ___) | | | (__| |__| (_| | || (_| | | (_) | (_| |
# |____/|_|  \___|\____\__,_|\__\__,_|_|\___/ \__, |
#                                             |___/

class SrcCatalog(phs.SrcCatalog):
    """Class for holding a catalog of celestial sources."""
    def get_jys(self, srcs=None):
        """Return list of fluxes of all src objects in catalog."""
        if srcs is None: srcs = self.keys()
        return np.array([self[s].get_jys() for s in srcs])
    def update_jys(self, afreqs):
        for s in self.keys(): self[s].update_jys(afreqs)

#  ____
# | __ )  ___  __ _ _ __ ___
# |  _ \ / _ \/ _` | '_ ` _ \
# | |_) |  __/ (_| | | | | | |
# |____/ \___|\__,_|_| |_| |_|

class Beam(phs.Beam):
    """Representation of a flat (gain=1) antenna beam pattern."""
    def response(self, xyz):
        """Return the (unity) beam response as a function of position."""
        x,y,z = np.array(xyz)
        return np.ones((self.afreqs.size, x.size))

class Beam2DGaussian(phs.Beam):
    """Representation of a 2D Gaussian beam pattern, with default setting for 
    a flat beam."""
    def __init__(self, freqs, xwidth=np.Inf, ywidth=np.Inf):
        """xwidth = angular width (radians) in EW direction
        ywidth = angular width (radians) in NS direction"""
        phs.Beam.__init__(self, freqs)
        self.xwidth, self.ywidth = xwidth, ywidth
    def response(self, xyz):
        """Return beam response across active band for specified topocentric 
        coordinates: (x=E,y=N,z=UP). x,y,z may be arrays of multiple 
        coordinates.  Returns 'x' linear polarization (rotate pi/2 for 'y')."""
        x,y,z = xyz
        x,y = np.arcsin(x)/self.xwidth, np.arcsin(y)/self.ywidth
        resp = np.sqrt(np.exp(-0.5*(x**2 + y**2)))
        resp = np.resize(resp, (self.afreqs.size, resp.size))
        return resp

class BeamPolynomial(phs.Beam):
    """Representation of a gaussian beam model whose width varies with azimuth
    angle and with frequency."""
    def __init__(self, freqs, poly_azfreq=np.array([[.5]])):
        """poly_azfreq = a 2D polynomial in cos(2*n*az) for first axis and 
        in freq**n for second axis."""
        self.poly = poly_azfreq
        phs.Beam.__init__(self, freqs)
        self.poly = poly_azfreq
        self._update_sigma()
    def select_chans(self, active_chans):
        """Select only enumerated channels to use for future calculations."""
        phs.Beam.select_chans(self, active_chans)
        self.update()
    def _update_sigma(self):
        f = np.resize(self.afreqs, (self.poly.shape[1], self.afreqs.size))
        f = f**np.array([range(self.poly.shape[1])]).transpose()
        self.sigma = np.dot(self.poly, f)
    def update(self):
        phs.Beam.update(self)
        self._update_sigma()
    def response(self, top):
        """Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y')."""
        az,alt = coord.top2azalt(top)
        zang = np.pi/2 - alt
        if zang.size == 1:
            zang = np.array([zang]); zang.shape = (1,)
            az = np.array([az]); az.shape = (1,)
        a = 2 * np.arange(self.poly.shape[0], dtype=np.float)
        a.shape = (1,) + a.shape; az.shape += (1,); zang.shape += (1,)
        a = np.cos(np.dot(az, a))
        a[:,0] = 0.5
        s = np.dot(a, self.sigma)
        return np.sqrt(np.exp(-(zang/s)**2)).transpose()

class BeamAlm(phs.Beam):
    """Representation of a beam model where each pointing has a response
    defined as a polynomial in frequency, and the spatial distributions of 
    these coefficients decomposed into spherical harmonics."""
    def __init__(self, freqs, lmax=8, mmax=8, deg=7, nside=64, coeffs={}):
        """lmax = maximum spherical harmonic term
        mmax = maximum spherical harmonic term in the z direction
        deg = order of polynomial to used for mapping response of each pointing
        nside = resolution of underlying HealpixMap to use
        coeffs = dictionary of polynomial term (integer) and corresponding Alm 
        coefficients (see healpix.py doc)."""
        phs.Beam.__init__(self, freqs)
        self.alm = [healpix.Alm(lmax,mmax) for i in range(deg+1)]
        self.hmap = [healpix.HealpixMap(nside,scheme='RING',interp=True)
            for a in self.alm]
        for c in coeffs:
            if c < len(self.alm): self.alm[-1-c].set_data(coeffs[c])
        self._update_hmap()
    def _update_hmap(self):
        for c,alm in enumerate(self.alm): self.hmap[c].from_alm(self.alm[c])
    def update(self):
        """Update beam model using new set of coefficients.
        coeffs = dictionary of polynomial term (integer) and corresponding Alm 
        coefficients (see healpix.py doc)."""
        phs.Beam.update(self)
        self._update_hmap()
    def response(self, top):
        """Return beam response across active band for specified topocentric 
        coordinates (x=E,y=N,z=UP). x,y,z may be multiple coordinates.  
        Returns 'x' pol (rotate pi/2 for 'y')."""
        top = [healpix.mk_arr(c, dtype=np.double) for c in top]
        px,wgts = self.hmap[0].crd2px(*top, **{'interpolate':1})
        poly = np.array([np.sum(h.map[px] * wgts, axis=-1) for h in self.hmap])
        rv = np.polyval(poly, np.reshape(self.afreqs, (self.afreqs.size, 1)))
        return rv

#     _          _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` |
#  / ___ \| | | | ||  __/ | | | | | | (_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_|

class Antenna(phs.Antenna):
    """Representation of physical location and beam pattern of individual 
    antenna in array."""
    def __init__(self, x, y, z, beam, phsoff=[0.,0.], bp_r=np.array([1]),
            bp_i=np.array([0]), amp=1, pointing=(0.,np.pi/2,0), **kwargs):
        """x,y z = antenna coordinates in equatorial (ns) coordinates
        beam = Beam object (implements response() function)
        phsoff = polynomial phase vs. frequency.  Phs term that is linear
                 with freq is often called 'delay'.
        bp_r = polynomial (in freq) modeling real component of passband
        bp_i = polynomial (in freq) modeling imaginary component of passband
        amp = overall multiplicative scaling of gain
        pointing = antenna pointing (az,alt).  Default is zenith."""
        phs.Antenna.__init__(self, x,y,z, beam=beam, phsoff=phsoff)
        self.set_pointing(*pointing)
        self.bp_r = bp_r
        self.bp_i = bp_i
        self.amp = amp
        self._update_gain()
    def _update_gain(self):
        bp = np.polyval(self.bp_r, self.beam.afreqs) + \
             1j*np.polyval(self.bp_i, self.beam.afreqs)
        self._gain = self.amp * bp
    def update(self):
        phs.Antenna.update(self)
        self._update_gain()
    def set_pointing(self, az=0, alt=np.pi/2, twist=0):
        """Set the antenna beam to point at (az, alt) with specified
        right-hand twist to polarizations.  Polarization y is assumed
        to be +pi/2 azimuth from pol x."""
        y, z = np.array([0,1,0]), np.array([0,0,1])
        # Twist is negative b/c we apply it to the coords, not the beam
        rot = coord.rot_m(-twist, z)
        rot = np.dot(rot, coord.rot_m(alt-np.pi/2, y))
        # NOTE the az, alt definition, az (clockwise around z = up, 0 at x axis = north), alt (from horizon), see also coord.azalt2top
        # rot = np.dot(rot, coord.rot_m(-az, z))
        rot = np.dot(rot, coord.rot_m(-(np.pi/2 - az), z))
        self.rot_pol_x = rot
        self.rot_pol_y = np.dot(coord.rot_m(-np.pi/2, z), rot)
    def bm_response(self, top, pol='x'):
        """Return response of beam for specified polarization."""
        top = np.array(top)
        top = {'x':np.dot(self.rot_pol_x, top), 
               'y':np.dot(self.rot_pol_y, top)}[pol]
        x,y,z = top
        return self.beam.response((x,y,z))
    def passband(self, conj=False):
        if conj: return np.conjugate(self._gain)
        else: return self._gain

#     _          _                            _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _   / \   _ __ _ __ __ _ _   _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` | / _ \ | '__| '__/ _` | | | |
#  / ___ \| | | | ||  __/ | | | | | | (_| |/ ___ \| |  | | | (_| | |_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_/_/   \_\_|  |_|  \__,_|\__, |
#                                                                 |___/ 

class AntennaArray(phs.AntennaArray):
    """Representation of location and time of observation, and response of
    array of antennas as function of pointing and frequency."""
    def __init__(self, *args, **kwargs):
        phs.AntennaArray.__init__(self, *args, **kwargs)
        self.active_pol = None
    def set_active_pol(self, pol):
        assert(pol in ('xx','yy','xy','yx')) # the only pols supported natively, for now
        self.active_pol = pol
    def get_active_pol(self):
        if self.active_pol is None: raise RuntimeError('No active polarization set (use AntennaArray.set_active_pol)')
        return self.active_pol
    def set_jultime(self, t=None):
        """Set current time as a Julian date."""
        phs.AntennaArray.set_jultime(self, t=t)
        self.eq2top_m = coord.eq2top_m(-self.sidereal_time(), self.lat)
        self._cache = None
    def passband(self, i, j):
        """Return the passband response of baseline i,j."""
        return self[j].passband() * self[i].passband(conj=True)
    def bm_response(self, i, j):
        """Return the beam response towards the cached source positions
        for baseline i,j with the specified polarization."""
        pol = self.get_active_pol()
        p1, p2 = pol[0], pol[-1]
        # Check that we have cached results needed.  If not, cache them.
        for c,p in zip([i,j], [p1,p2]):
            if not self._cache.has_key(c): self._cache[c] = {}
            if not self._cache[c].has_key(p):
                x,y,z = self._cache['s_top']
                resp = self[c].bm_response((x,y,z), pol=p).transpose()
                self._cache[c][p] = resp
        return self._cache[j][p2] * np.conjugate(self._cache[i][p1])
    def sim_cache(self, s_eqs, jys=np.array([1.]), mfreqs=0.150,
            ionrefs=(0.,0.), srcshapes=(0,0,0)):
        """Cache intermediate computations given catalog information to speed
        simulation for multiple baselines.  For efficiency, should only be 
        called once per time setting.  MUST be called before sim().
        s_eqs = array of equatorial vectors for all celestial sources
        jys = array of janskies vs. freq for all celestial sources
        mfreqs = array of frequencies where ionrefs were measured
        ionrefs = (dra,ddec), angular offsets in radians for ra/dec at the
            frequency `mfreq'.
        srcshapes = (a1,a2,th) where a1,a2 are angular sizes along the 
            semimajor, semiminor axes, and th is the angle (in radians) of
            the semimajor axis from E."""
        # Get topocentric coordinates of all srcs
        src_top = np.dot(self.eq2top_m, s_eqs)
        # Throw out everything that is below the horizon
        valid = src_top[2,:] > 0
        if np.all(valid == 0): self._cache = {}
        else:
            jys = jys.compress(valid, axis=0)
            try:
                mfreqs = mfreqs.compress(valid)
                mfreqs.shape = (mfreqs.size, 1)
            except(AttributeError): pass
            a1,a2,th = srcshapes
            try:
                a1 = a1.compress(valid)
                a2 = a2.compress(valid)
                th = th.compress(valid)
            except(AttributeError): pass
            dra,ddec = ionrefs
            try:
                dra = dra.compress(valid)
                ddec = ddec.compress(valid)
            except(AttributeError): pass
            src_top = src_top.compress(valid, axis=1)
            s_eqs = s_eqs.compress(valid, axis=1)
            # Get src fluxes vs. freq
            self._cache = {
                'jys':   jys,
                'mfreq': mfreqs,
                's_eqs': s_eqs,
                's_top': src_top,
                's_shp': (a1,a2,th),
                'i_ref': (dra,ddec),
            }
    def sim(self, i, j):
        """Simulate visibilites for the specified (i,j) baseline and 
        polarization (set with AntennaArray.set_active_pol).  sim_cache() 
        must be called at each time step before this returns valid results."""
        if self._cache is None:
            raise RuntimeError('sim_cache() must be called before the first sim() call at each time step.')
        elif self._cache == {}:
            return np.zeros_like(self.passband(i,j))
        s_eqs = self._cache['s_eqs']
        u,v,w = self.gen_uvw(i, j, src=s_eqs)
        I_sf = self._cache['jys']
        Gij_sf = self.passband(i,j)
        Bij_sf = self.bm_response(i,j)
        if len(Bij_sf.shape) == 2: Gij_sf = np.reshape(Gij_sf, (1, Gij_sf.size))
        # Get the phase of each src vs. freq, also does resolution effects
        E_sf = np.conjugate(self.gen_phs(s_eqs, i, j, mfreq=self._cache['mfreq'],
            srcshape=self._cache['s_shp'], ionref=self._cache['i_ref'],
            resolve_src=True))
        try: E_sf.shape = I_sf.shape
        except(AttributeError): pass
        # Combine and sum over sources
        GBIE_sf = Gij_sf * Bij_sf * I_sf * E_sf
        Vij_f = GBIE_sf.sum(axis=0)
        return Vij_f



#!/usr/bin/env python

"""
Do a one iteration delay-solving algorithm.
"""

import aipy as a
import numpy as np
from matplotlib import pylab as pl
import optparse,sys,os
from time import time
from glob import glob

o = optparse.OptionParser()
o.set_description(__doc__)
o.set_usage('find_delay.py [options] *.uv.src')
o.add_option('-C','--cal',dest='cal',type='str',help='Cal file to use.')
o.add_option('-f','--flag',dest='flag',type='str',help='Antennae to flag. Ex. 1,2,3')
o.add_option('-p','--plots',dest='plots',default=None,action='store_true',help='Plots?')
opts,args = o.parse_args(sys.argv[1:])

t0 = time()
if opts.cal != None:
    exec('from %s import prms'%opts.cal)
    old_delays = prms['delays'].copy()
    for ant in old_delays:
        if type(old_delays[ant]) is not dict:
            old_delays[ant] = {'x': prms['delays'][ant]}

#Parse Arguments
FlagAnt = []
if opts.flag:
    for antstr in opts.flag.split(','):
        FlagAnt.append(int(antstr))

#Define a function to give me a cleaned delay.
def clean_delay(d):
    d = d.squeeze()
    d = np.fft.ifft(d)
    samples = np.logical_not(f).astype(float)
    ker = np.fft.ifft(samples)
    gain = np.sqrt(np.mean(samples**2))
    d,info = a.deconv.clean(d,ker,tol=1e-6)
    d += info['res']/gain
    d = np.fft.fftshift(d)
    return d

def length(A):
    return np.sqrt(np.dot(A,A))

aa = a.cal.get_aa(opts.cal,np.array([0.15]))

#Read in Data
DD = {}
bl_len = {}
AntNos = []
for uvfile in args:
    print 'Reading',uvfile
    uv = a.miriad.UV(uvfile)
    for (uvw,t,(i,j)),d,f in uv.all(raw=True):
        if i == j: continue
        if not (i,j) in bl_len.keys(): bl_len[(i,j)] = length(aa.get_baseline(i,j))
        if bl_len[(i,j)]*0.15 <= 30.: continue
        pol = a.miriad.pol2str[uv['pol']]
        if pol == 'xy' : continue
        if pol == 'yx' : continue
        if i in FlagAnt: continue
        if j in FlagAnt: continue
        if not pol in DD.keys(): 
            DD[pol] = {}
        if not (i,j) in DD[pol].keys(): DD[pol][(i,j)] = np.zeros(uv['nchan'],dtype=np.complex)
        #Delay transform
        d = np.ma.array(d)
        DD[pol][(i,j)] += d
        if not i in AntNos: AntNos.append(i)
        if not j in AntNos: AntNos.append(j)
pols = DD.keys()
AntNos.sort()
pols.sort()
delays = np.linspace(-0.5/uv['sdf'],0.5/uv['sdf'],uv['nchan'])
del(uv)

Npol = len(DD.keys())
Nant = len(AntNos)
Nbl = int(Nant*(Nant-1)/2)

#Find the peak of each delay spectrum
D_ij = {}
W_ij = {}
for pol in pols:
    D_ij[pol] = {}
    W_ij[pol] = {}
    for bl in DD[pol]:
        DD[pol][bl] = np.abs(clean_delay(DD[pol][bl]))
        DD[pol][bl] /= DD[pol][bl].max()
        #multiply by some prior.
        for i,x in enumerate(DD[pol][bl]):
            if x == 1: D_ij[pol][bl] = delays[i]
        #W_ij[pol][bl] = 1./(np.std(DD[pol][bl])**2) 
        W_ij[pol][bl] = 1. 

#Construct matrices to do linear algebra and shit.

Tau = {}
for pol in pols:
    A = np.zeros((Nbl+1,Nant))
    W = np.zeros((Nbl+1,Nbl+1))
    b = np.zeros(Nbl+1)
        
    #populatin the countryside, populatin the villages!
    index = 0
    for i,ant1 in enumerate(AntNos):
        for j,ant2 in enumerate(AntNos):
            if not (ant1,ant2) in D_ij[pol].keys(): continue
            #if j <= i: continue
            b[index] = D_ij[pol][(ant1,ant2)]
            A[index,i],A[index,j] = 1.,-1.    
            W[index,index] = W_ij[pol][(ant1,ant2)]
            index += 1
    #Set Constraints
    W /= np.sum(W)
        
    Tau[pol] = np.dot(np.linalg.inv(np.dot(np.dot(A.T,W),A)),np.dot(np.dot(A.T,W),b))
    Tau[pol] -= np.mean(Tau[pol])

#Print things so you can just copy/paste into a cal file.
if not opts.cal is None:
    for i,ant in enumerate(AntNos):
        calstr = str(ant)+' : { '
        for pol in pols:
            calstr +="'"+pol[0]+"'"+' : '
            Tau[pol][i] += old_delays[ant][pol[0]]
            calstr += str(Tau[pol][i])
            if pol != pols[-1]: calstr += ', '
        calstr += ' }'
        if i != len(AntNos)-1: calstr += ','
        print calstr
else:
    for i,ant in enumerate(AntNos):
        calstr = str(ant)+' : { '
        for pol in pols:
            calstr +="'"+pol[0]+"'"+' : '
            calstr += str(Tau[pol][i])
            if pol != pols[-1]: calstr += ', '
        calstr += ' }'
        if i != len(AntNos)-1: calstr += ','
        print calstr

print 'Computing time =', time() - t0,'s for',Nbl,'baselines.'


if opts.plots:
    print 'Generating plots.... this may take a while...'
    figcnt = 0
    AntNos = np.array(AntNos) 
    
    def issubplotedge(rows,cols,index):
        inds = np.arange(1,cols*rows+1,1)
        inds.shape = (rows,cols)
        left = not np.argwhere(inds==index).squeeze()[1]
        bottom = not (np.argwhere(inds==index).squeeze()[0]+1)%rows
        return left,bottom
    
    for pol in pols:
        pl.figure(figcnt)
        
        nplots = len(DD[pol].keys())
        rows = Nant
        cols = Nant
        bls = DD[pol].keys()
        bls.sort()

        for plno,bl in enumerate(bls):
            #Make the plots
            i,j = np.argwhere(AntNos==bl[0]).squeeze(),np.argwhere(AntNos==bl[1]).squeeze()
            plindex = i*Nant+j
            
            ax = pl.subplot(rows,cols,plindex)
           
            pl.vlines(D_ij[pol][bl]/bl_len[bl],0,1.1,color='k')
            pl.vlines((Tau[pol][i]-Tau[pol][j])/bl_len[bl],0,1.1,color='r')
            pl.vlines((-1,1),0,1.1,linestyles='dotted',color='k')
            pl.plot(delays/bl_len[bl],DD[pol][bl],'b')
           
            pl.xlim([-1.5,1.5])
            pl.ylim([0,1.1])
            pl.yticks([])
            pl.xticks([])

            if i == 0: pl.text(0.5,1.2,str(AntNos[j]),transform=ax.transAxes)
            if j == Nant-1: pl.text(1.2,0.5,str(AntNos[i]),transform=ax.transAxes)
    
        #Give a sample plot
        pl.subplot(337)
        bl = bls[0]
        i,j = np.argwhere(AntNos==bl[0]).squeeze(),np.argwhere(AntNos==bl[1]).squeeze()
        pl.vlines(D_ij[pol][bl]/bl_len[bl],0,1.1,color='k')
        pl.vlines((Tau[pol][i]-Tau[pol][j])/bl_len[bl],0,1.1,color='r')
        pl.vlines((-1,1),0,1.1,linestyles='dotted',color='k')
        pl.plot(delays/bl_len[bl],DD[pol][bl],'b')
        pl.title('Sample baseline, %d_%d'%bl)
        pl.xlim([-1.5,1.5])
        pl.ylim([0,1.1])
        pl.xlabel('Delay, baseline lenghth.')
        pl.ylabel('Amplitude')

        pl.subplots_adjust(wspace=0,hspace=0,top=0.95,bottom=0.05)
        pl.draw()
        figcnt += 1
        
    #Plot a matrix of all delays
    for pol in pols:
        pl.figure(figcnt)
        DDD = np.zeros((Nant,Nant))
        for i,ant1 in enumerate(AntNos):
            for j,ant2 in enumerate(AntNos):
                if not (ant1,ant2) in D_ij[pol].keys(): continue
                DDD[i,j] = D_ij[pol][(ant1,ant2)]
        DDD = np.ma.array(DDD,mask=np.where(DDD==0,1,0))
        pl.matshow(DDD,aspect='auto')
        pl.colorbar()
        pl.draw()
        figcnt += 1
        
    
    #Delay vs. Antenna
    pl.figure(figcnt)
    for pol in pols:
        pl.plot(AntNos,Tau[pol],'o',label=pol[0])
        pl.xlabel('Antenna Number')
        pl.ylabel('Delay (ns)')
    pl.legend()
    pl.draw()

    pl.show()



#! /usr/bin/env python
"""
A script for filtering using a delay/delay-rate transform.  If a source
is specified, will remove/extract that source.  If none is specified,
will filter/extract in absolute terms.
"""

import aipy as a, numpy as np, os, sys, optparse, math

def gen_skypass_delay(aa, sdf, nchan, pol, max_bl_frac=1.5):
    aa.set_active_pol(pol)
    bin_dly = 1. / (sdf * nchan)
    filters = {}
    for i in range(len(aa.ants)):
      for j in range(len(aa.ants)):
        if j <= i: continue
        bl = aa.ij2bl(i,j)
        max_bl = aa.get_baseline(i,j)
        max_bl = max_bl_frac * np.sqrt(np.dot(max_bl, max_bl))
        dly,off = aa.get_phs_offset(i,j)[-2:]
        uthresh, lthresh = (dly + max_bl)/bin_dly + 1, (dly - max_bl)/bin_dly
        uthresh, lthresh = int(np.round(uthresh)), int(np.round(lthresh))
        filters[bl] = (uthresh,lthresh)
    return filters

o = optparse.OptionParser()
o.set_usage('filter_src.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, src=True, cal=True)
o.add_option('-p','--pol',dest='pol',default='xx,xy,yx,yy',help='Polarizations to use.')
o.add_option('-r', '--drw', dest='drw', type=int, default=-1,
    help='The number of delay-rate bins to null.  Default is -1 = no fringe filtering.')
o.add_option('-d', '--dw', dest='dw', type=int, default=5,
    help='The number of delay bins to null. If -1, uses baseline lengths to generate a sky-pass filter.')
o.add_option('-P','--passband', dest='passband', action='store_true',
    help='Divide by the model passband before transforming.')
o.add_option('-e', '--extract', dest='extract', action='store_true',
    help='Extract the source instead of removing it.')
o.add_option('--clean', dest='clean', type='float', default=1e-3,
    help='Deconvolve delay-domain data by the response that results from flagged data.  Specify a tolerance for termination (usually 1e-2 or 1e-3).')
opts, args = o.parse_args(sys.argv[1:])

pols = opts.pol.split(',')

uv = a.miriad.UV(args[0])
if not opts.src is None:
    aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
    srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
    cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
    src = cat[opts.src]
else:
    aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
    src = None
    filters = {}
    for pol in pols:
        filters[pol] = gen_skypass_delay(aa, uv['sdf'], uv['nchan'],pol)
del(uv)

for uvfile in args:
    phs_dat = {}
    if src is None: uvofile = uvfile + 'f'
    elif opts.extract: uvofile = uvfile+'.e_'+opts.src
    else: uvofile = uvfile+'.f_'+opts.src
    print uvfile,'->',uvofile
    if os.path.exists(uvofile):
        print uvofile, 'exists, skipping.'
        continue
    uvi = a.miriad.UV(uvfile)

    curtime = None
    print '    Performing delay transform and cleaning...'
    src_up = False
    for (uvw,t,(i,j)),d,f in uvi.all(raw=True):
        pol = a.miriad.pol2str[uvi['pol']]
        if not pol in pols: continue
        aa.set_active_pol(pol)
        if i == j: continue
        if not pol in phs_dat.keys(): phs_dat[pol] = {}
        if curtime != t:
            curtime = t
            if not src is None:
                aa.set_jultime(t)
                cat.compute(aa)
        bl = a.miriad.ij2bl(i,j)
        try:
            flags = np.logical_not(f).astype(np.float)
            # Check first if source is up
            if not src is None:
                d = aa.phs2src(d, src, i, j)
            # Put passband into kernel, where it gets divided out of the data
            if opts.passband: flags *= aa.passband(i,j)
            gain = np.sqrt(np.average(flags**2))
            ker = np.fft.ifft(flags)
            d = np.where(f, 0, d)
            src_up = True
            d = np.fft.ifft(d)
            if not np.all(d == 0):
                d, info = a.deconv.clean(d, ker, tol=opts.clean)
                d += info['res'] / gain
        except(a.phs.PointingError): d = np.zeros_like(d)
        try: phs_dat[pol][bl].append(d)
        except(KeyError): phs_dat[pol][bl] = [d]

    if opts.drw != -1:
        print '    Performing delay-rate transform and cleaning...'
    for pol in phs_dat:
        for bl in phs_dat[pol]:
            d = np.array(phs_dat[pol][bl])
            if not src_up:
                phs_dat[pol][bl] = d
                continue
            # create some padding data on either end to mitigate wrap-around
            # effects of a delay-rate filter
            padlen = math.ceil(d.shape[0] * .1)
            d = np.concatenate([np.flipud(d[:padlen]), d, np.flipud(d[-padlen:])])
            if opts.drw != -1:
                flags = np.where(d[:,0] != 0, 1., 0.)
                gain = np.sqrt(np.average(flags**2))
                ker = np.fft.ifft(flags)
                d = np.fft.ifft(d, axis=0)
                for chan in range(d.shape[1]):
                    d[:,chan],info = a.deconv.clean(d[:,chan],ker,tol=opts.clean)
                    d[:,chan] += info['res'] / gain
                x1, x2 = opts.drw, -opts.drw+1
                if x2 == 0: x2 = d.shape[0]
                d[x1:x2,:] = 0
            if opts.dw != -1: 
                y1, y2 = opts.dw, -opts.dw
                if y2 == 0: y2 = d.shape[1]
            else: y1, y2 = filters[pol][bl]
            d[:,y1:y2] = 0
            if opts.drw != -1: d = np.fft.fft(d, axis=0)
            d = np.fft.fft(d, axis=1)
            # unpad the data
            d = d[padlen:-padlen]
            phs_dat[pol][bl] = d

    cnt = {}
    for pol in phs_dat:
        cnt[pol] = {}
        for bl in phs_dat[pol]: cnt[pol][bl] = 0

    # Generate a pipe for removing average phase bias from data
    curtime = None
    def rm_mfunc(uv, p, d):
        global curtime
        uvw, t, (i,j) = p
        if i == j: return p, d
        pol = a.miriad.pol2str[uv['pol']]
        aa.set_active_pol(pol)
        bl = a.miriad.ij2bl(i,j)
        data = phs_dat[pol][bl][cnt[pol][bl],:]
        if not src is None:
            if curtime != t:
                curtime = t
                aa.set_jultime(t)
                cat.compute(aa)
            try: data = aa.unphs2src(data, src, i, j)
            except(a.phs.PointingError): data *= 0
        # Remultiply by passband b/c the kernel divided it out
        if opts.passband:
            data *= aa.passband(i,j)
        cnt[pol][bl] += 1
        if opts.extract: return p, np.ma.array(data, mask=d.mask)
        else: return p, d - data

    print '    Writing out filtered data...'
    uvi.rewind()
    uvo = a.miriad.UV(uvofile, status='new')
    uvo.init_from_uv(uvi)
    # Apply the pipe to the data
    uvo.pipe(uvi, mfunc=rm_mfunc, append2hist='FILTER_SRC: src=%s drw=%d dw=%d extract=%s clean=%f passband=%s beam=%s\n' % (opts.src, opts.drw, opts.dw, opts.extract, opts.clean, opts.passband, False))



#! /usr/bin/env python
"""
A script for listing sources (and optionally source parameters) from
catalogs.
"""
import aipy as a, ephem, sys, optparse,logging

o = optparse.OptionParser()
o.set_usage('srclist.py [options]')
o.set_description(__doc__)
a.scripting.add_standard_options(o, src=True, cal=True)
o.add_option('-P','--prms', dest='prms', 
    help='A comma-delimited list of parameters to print for each source.  Can be "*" to list all parameters.  If no parameters are specified, a list of the selected sources is printed.')
o.add_option('-x','--exclude', dest='exclude', 
    help='A string that is parsed similarly to "srcs", but excludes sources that may otherwise have been selected by the "srcs" parameter.')
o.add_option('-c','--cen', dest='centers', 
    help='A string that is parsed similarly to "srcs", but is used along with --sep to select sources near specified locations.')
o.add_option('-j','--juldate', dest='juldate', type='float',
    help='A Julian Date to use for placing sources.')
o.add_option('--ra',dest='ra_rng', 
    help='A range RA1_RA2 of right-ascensions to select for.  Default: None.')
o.add_option('--dec',dest='dec_rng',
    help='A range DEC1_DEC2 of declinations to select for.  Default: None.')
o.add_option('--sep',dest='sep', type='float', 
    help='Include areas within the specified angular separation (in degrees) of any sources listed in --src.')
o.add_option('--divstr', dest='divstr', default=' ',
    help='Divider string to use between source names when printing.  Default is " ".')
o.add_option('-v',dest='verb',action='store_true',
    help="Print more")
o.add_option('--fitprms',action='store_true',
    help="Print as prm inputs for fitmdl")
opts,args = o.parse_args(sys.argv[1:])

srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
if opts.verb:
    logging.basicConfig(level=logging.DEBUG)
else:
     logging.basicConfig(level=logging.WARNING)
log = logging.getLogger('srclist')

if opts.cal != None:
    cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
else:
    cat = a.src.get_catalog(srclist, cutoff, catalogs)

if opts.exclude != None:
    xlist,xoff,catalogs = a.scripting.parse_srcs(opts.exclude, opts.cat)
    if opts.cal != None:
        xcat = a.cal.get_catalog(opts.cal, xlist, xoff, catalogs)
    else:
        xcat = a.src.get_catalog(xlist, xoff, catalogs)
else: xcat = {}

if opts.centers != None:
    assert(opts.sep != None)
    clist,coff,catalogs = a.scripting.parse_srcs(opts.centers, opts.cat)
    if opts.cal != None:
        ccat = a.cal.get_catalog(opts.cal, clist, coff, catalogs)
    else:
        ccat = a.src.get_catalog(clist, coff, catalogs)
else: ccat = {}
    
if opts.juldate is None: date = ephem.J2000
else: 
    date = a.phs.juldate2ephem(opts.juldate)
    aa = a.scripting.get_null_aa()
    aa.set_jultime(opts.juldate)

for c in [cat, xcat, ccat]:
    for s in c.keys():
        try: ephem.FixedBody.compute(c[s], date)
        except(TypeError):
            if opts.juldate is None: del(c[s])
            else: c[s].compute(aa)

srcs = cat.keys()
srcs = [s for s in srcs if s not in xcat]
if opts.sep != None:
    nsrcs = []
    for s1 in srcs:
        for s2 in ccat.keys():
            if ephem.separation(cat[s1], ccat[s2]) <= opts.sep * a.img.deg2rad:
                nsrcs.append(s1)
                break
    srcs = nsrcs
    

if opts.ra_rng != None:
    ra1,ra2 = map(ephem.hours, opts.ra_rng.split('_'))
    if ra1 < ra2:
        srcs = [s for s in srcs if (cat[s].ra > ra1 and cat[s].ra < ra2)]
    else:
        srcs = [s for s in srcs if (cat[s].ra > ra1 or cat[s].ra < ra2)]

if opts.dec_rng != None:
    dec1,dec2 = map(ephem.degrees, opts.dec_rng.split('_'))
    if dec1 < dec2:
        srcs = [s for s in srcs if (cat[s].dec > dec1 and cat[s].dec < dec2)]
    else:
        srcs = [s for s in srcs if (cat[s].dec > dec1 or cat[s].dec < dec2)]

# We're done selecting sources now.  Time to print information
srcs.sort()
if opts.prms == None:
    print opts.divstr.join(srcs)
else:
    if not opts.fitprms:
        prms = opts.prms.split(',')
        for s in srcs:
            p = cat.get_params({s:prms})
            if p[s].has_key('ra'):  p[s]['ra'] = ephem.hours(p[s]['ra'])
            if p[s].has_key('dec'): p[s]['dec'] = ephem.degrees(p[s]['dec'])
            a.fit.print_params(p)
    else:
        prms = opts.prms.split(',')
        outstring = ''
        scount =0
        snum = len(srcs)
        for s in srcs:
            scount +=1
            outstring += s+"=("
            p = cat.get_params({s:prms})
            pcnt = len(p[s].keys())
            for i,prm in enumerate(p[s].keys()):
                if i<pcnt-1: outstring += prm+"/"
                else: outstring += prm
            if scount<snum-1:outstring += "),"
            else: outstring += ")"
        print outstring



"""
Module for detecting and flaging RFI related effects.
"""

import numpy as np, optimize

def gaussian(amp, sig, off, x):
    """Generate gaussian value at x given amplitude, sigma, and x offset."""
    return amp * np.exp(-((x - off)/sig)**2)

def fit_gaussian(xs, ys):
    """Fit a gaussian (returning amplitude, sigma, offset) to a set of x
    values and their corresponding y values."""
    def fitfunc(args):
        amp, sig, off = args
        sum = ((ys - gaussian(amp, sig, off, xs))**2).sum()
        return sum
    return optimize.fmin(fitfunc, (ys.sum(), 50, ys.argmax()), disp=0)

def gen_rfi_thresh(data, nsig=2, cnt_per_bin=1000):
    """Generate a threshold at nsig times the standard deviation (above,below)
    the mean, outside of which data is considered rfi."""
    try: data = data.compressed()
    except(AttributeError): pass
    data = np.log10(np.abs(data))
    try: h, bvals = np.histogram(data, bins=data.size/cnt_per_bin)
    except(ValueError): return None, None
    if h.size == 0: return None, None
    # Fit a gaussian to histogram (better than just std-dev of data)
    amp, sig, off = fit_gaussian(np.arange(h.size), h)
    sig = abs(sig)
    hi_thresh = np.clip(np.round(off + nsig*sig), 0, len(bvals)-1)
    lo_thresh = np.clip(np.round(off - nsig*sig), 0, len(bvals)-1)
    return 10**bvals[hi_thresh], 10**bvals[lo_thresh]

def flag_by_int(preflagged_auto, nsig=1, raw=False):
    """Flag rfi for an autocorrelation.  Iteratively removes outliers and
    fits a smooth passband, then uses smooth passband to remove outliers
    (both positive and negative).  If 'raw' is specified, no smooth function
    is removed."""
    pwr_vs_t = np.ma.average(np.abs(preflagged_auto), axis=1)
    mask = pwr_vs_t.mask
    if raw:
        spikey_pwr_vs_t = pwr_vs_t
    else: 
        spikey_pwr_vs_t = np.abs(pwr_vs_t - remove_spikes(pwr_vs_t, mask.copy()))
    hi_thr, lo_thr = gen_rfi_thresh(spikey_pwr_vs_t, cnt_per_bin=20, nsig=nsig)
    spikey_pwr_vs_t = spikey_pwr_vs_t.filled(hi_thr)
    mask = spikey_pwr_vs_t >= hi_thr
    return mask.astype(np.int)

def remove_spikes(data, mask=None, order=6, iter=3, return_poly=False):
    """Iteratively fits a smooth function by removing outliers and fitting
    a polynomial, then using the polynomial to remove other outliers."""
    xs = np.arange(data.size)
    if mask is None or mask.size != data.size:
        mask = np.zeros(data.shape, dtype=np.bool)
    im = np.logical_not(mask)
    nxs = xs.compress(im)
    ndata = data.compress(im)
    if len(nxs) != 0: p = np.polyfit(nxs, ndata, deg=order)
    else: p = np.polyfit(xs, data, deg=order)
    if iter != 0:
        residual = np.abs(ndata - np.polyval(p, nxs))
        sig = np.sqrt(np.average(residual**2))
        mask.put(np.where(residual > sig), 1)
        p = remove_spikes(data, mask=mask, order=order,
            iter=iter-1, return_poly=True)
    if return_poly: return p
    else: return np.polyval(p, xs)




#! /usr/bin/env python
"""
This is a general-purpose script for making images from MIRIAD UV files.  Data
(optionally selected for baseline, channel) are read from the file, phased
to a provided position, normalized for passband/primary beam effects, gridded
to a UV matrix, and imaged
"""

import aipy as a, numpy as np, sys, optparse, ephem, os

o = optparse.OptionParser()
o.set_usage('mk_img.py [options] *.uv')
o.set_description(__doc__)
a.scripting.add_standard_options(o, ant=True, pol=True, chan=True, cal=True,
    src=True, dec=True)
o.add_option('-o', '--output', dest='output', default='dim,dbm',
    help='Comma delimited list of data to generate FITS files for.  Can be: dim (dirty image), dbm (dirty beam), uvs (uv sampling), or bms (beam sampling).  Default is dim,dbm.')
o.add_option('--list_facets', dest='list_facets', action='store_true',
    help='List the coordinates of all the pointings that will be used.')
o.add_option('--facets', dest='facets', type='int', default=200,
    help='If no src is provided, facet the sphere into this many pointings for making a map.  Default 200.')
o.add_option('--snap', dest='snap', type='int', default=-1,
    help='Number of integrations to use in "snapshot" images.  Default is to not do snapshoting (i.e. all integrations go into one image).')
o.add_option('--cnt', dest='cnt', type='int', default=0,
    help='Start counting output images from this number.  Default 0.')
o.add_option('--fmt', dest='fmt', default='im%04d',
    help='A format string for counting successive images written to files.  Default is im%04d (i.e. im0001).')
o.add_option('--skip_phs', dest='skip_phs', action='store_true',
    help='Do not phase visibilities before gridding.')
o.add_option('--zen_phased', dest='zen_phased', action='store_true',
    help='If data have already have already been phased to zenith, set this flag.')
o.add_option('--skip_amp', dest='skip_amp', action='store_true',
    help='Do not use amplitude information to normalize visibilities.')
o.add_option('--skip_bm', dest='skip_bm', action='store_true',
    help='Do not weight visibilities by the strength of the primary beam.')
o.add_option('--skip', dest='skip', type='int', default=0,
    help='Skip this many pointings before starting.  Useful in conjungtion with --cnt for resuming.')
o.add_option('--size', dest='size', type='int', default=300,
    help='Size of maximum UV baseline.')
o.add_option('--res', dest='res', type='float', default=0.5,
    help='Resolution of UV matrix.')
o.add_option('--no_w', dest='no_w', action='store_true',
    help="Don't use W projection.")
o.add_option('--wres', dest='wres', type='float', default=0.5,
    help="W-Plane projection resolution.  Default 0.5")
o.add_option('--altmin', dest='altmin', type='float', default=0,
    help="Minimum allowed altitude for pointing, in degrees.  When phase center is lower than this altitude, data is omitted.  Default is 0.")
o.add_option('--minuv', dest='minuv', type='float', default=0,
    help="Minimum distance from the origin in the UV plane (in wavelengths) for a baseline to be included.  Default is 0.")
o.add_option('--buf_thresh', dest='buf_thresh', default=2e6, type='float',
    help='Maximum amount of data to buffer before gridding.  Excessive gridding takes performance hit, but if buffer exceeds memory available... ouch.')
opts, args = o.parse_args(sys.argv[1:])

# Parse command-line options
uv = a.miriad.UV(args[0])
(j,t,j),j = uv.read()
chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
a.scripting.uv_selector(uv, opts.ant, opts.pol)
aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
aa.select_chans(chans)
aa.set_active_pol(opts.pol)
afreqs = aa[0].beam.afreqs
cfreq = np.average(afreqs)
aa.set_jultime(t)
del(uv)
outputs = opts.output.split(',')

# Get all sources that will be used as phase centers.  If no sources are
# specified, define phase centers for faceting a sphere.
if opts.src == 'zen':
    srcs = [a.phs.RadioFixedBody(aa.sidereal_time(), aa.lat, name='zen')]
    cat = a.phs.SrcCatalog(srcs)
elif not opts.src is None: 
    srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
    cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
else:
    ras,decs = a.map.facet_centers(opts.facets, ncrd=2)
    srcs = [a.phs.RadioFixedBody(ra,dec,name=str(i)) 
        for i,(ra,dec) in enumerate(zip(ras,decs))]
    cat = a.phs.SrcCatalog(srcs)

if opts.list_facets:
    cat.compute(aa)
    srcs = cat.keys(); srcs.sort()
    for cnt, src in enumerate(cat.values()):
        cen = ephem.Equatorial(src.ra, src.dec, epoch=aa.epoch)
        cen = ephem.Equatorial(cen, epoch=ephem.J2000)
        print '# %3d >  RA=%s  DEC=%s  (%f, %f in deg)' % \
            (cnt, cen.ra, cen.dec, 
            a.img.rad2deg*cen.ra, a.img.rad2deg*cen.dec)

# Generate the image object that will be used.
us,vs,ws,ds,wgts = [],[],[],[],[]
if opts.no_w:
    im = a.img.Img(opts.size, opts.res, mf_order=0)
else:
    im = a.img.ImgW(opts.size, opts.res, mf_order=0, wres=opts.wres)
L,M = im.get_LM()
DIM = int(opts.size/opts.res)
n_ints = 0

#print 'Calculating image of primary beam'
#top = im.get_eq(0, aa.lat)
#mask = top[0].mask
#m = a.coord.eq2top_m(0, aa.lat)
#top = top.transpose([1,0,2])
#x,y,z = np.dot(m, top)
#aa.select_chans([120])
#d = aa.ants[0].bm_response((x.flatten(),y.flatten(),z.flatten()), pol='y')[0]**2
#aa.select_chans(chans)
#d.shape = (DIM,DIM)
#bm_im = np.where(mask, 0, d)
#print 'done'

# Define a quick function writing an image to a FITS file
def fname(ftag, cnt): return '%s.%s.fits' % (opts.fmt % cnt, ftag)
def to_fits(ftag,i,src,cnt,history=''):
    filename = fname(ftag,cnt)
    print 'Saving data to', filename
    while len(i.shape) < 4: i.shape = i.shape + (1,)
    cen = ephem.Equatorial(src.ra, src.dec, epoch=aa.epoch)
    # We precess the coordinates of the center of the image here to
    # J2000, just to have a well-defined epoch for them.  For image coords to
    # be accurately reconstructed, precession needs to be applied per pixel
    # and not just per phase-center because ra/dec axes aren't necessarily
    # aligned between epochs.  When reading these images, to be 100% accurate,
    # one should precess the ra/dec coordinates back to the date of the
    # observation, infer the coordinates of all the pixels, and then
    # precess the coordinates for each pixel independently.
    cen = ephem.Equatorial(cen, epoch=ephem.J2000)
    a.img.to_fits(filename, i, clobber=True,
        object=src.src_name, obs_date=str(aa.date),
        ra=cen.ra*a.img.rad2deg, dec=cen.dec*a.img.rad2deg, epoch=2000.,
        d_ra=L[-1,-1]*a.img.rad2deg, d_dec=M[1,1]*a.img.rad2deg,
        freq=np.average(aa[0].beam.afreqs),history=history)

def grid_it(im,us,vs,ws,ds,wgts):
    #print 'Gridding %d integrations' % n_ints
    sys.stdout.write('|'); sys.stdout.flush()
    if len(ds) == 0: raise ValueError('No data to use.')
    ds,wgts = np.concatenate(ds), np.concatenate(wgts).flatten()
    us,vs,ws = np.concatenate(us), np.concatenate(vs), np.concatenate(ws)
    # Grid data into UV matrix
    (us,vs,ws),ds,wgts = im.append_hermitian((us,vs,ws),ds,wgts)
    im.put((us,vs,ws), ds, wgts)
    #im.put((us,vs,ws), ds, wgts, invker2=bm_im)

def img_it(im):
    global n_ints
    #print 'Imaging with %d integrations' % n_ints
    n_ints = 0
    # Form dirty images/beams
    uvs = a.img.recenter(np.abs(im.uv).astype(np.float), (DIM/2,DIM/2))
    bms = a.img.recenter(np.abs(im.bm[0]).astype(np.float), (DIM/2,DIM/2))
    dim = im.image((DIM/2, DIM/2))
    dbm = im.bm_image(term=0, center=(DIM/2,DIM/2))
    return uvs,bms, dim,dbm

# Loop through all specified sources, generating images
imgcnt = opts.cnt
for srccnt, s in enumerate(cat.values()):
    if srccnt < opts.skip: continue
    s.compute(aa)
    print '%d / %d' % (srccnt + 1, len(cat.values()))
    print 'Pointing (ra, dec):', s.ra, s.dec
    src = a.fit.SrcCatalog([s])
    # Gather data
    snapcnt,curtime = 0, None
    # Read each file
    for filename in args:
      sys.stdout.write('.'); sys.stdout.flush()
      uv = a.miriad.UV(filename)
      a.scripting.uv_selector(uv, opts.ant, opts.pol)
      uv.select('decimate', opts.decimate, opts.decphs)
      # Read all data from each file
      for (crd,t,(i,j)),d,f in uv.all(raw=True):
          pol = a.miriad.pol2str[uv['pol']]
          history = uv['history']
          history = history +  sys.argv[0].split('/')[-1].strip()+' ' + ' '.join(sys.argv[1:])
          if curtime != t:
              # Make snapshot images (if specified)
              if opts.snap > 0:
                  snapcnt = (snapcnt + 1) % opts.snap
                  if snapcnt == 0:
                      if curtime != None:
                          try:
                              grid_it(im,us,vs,ws,ds,wgts)
                              uvs,bms,dim,dbm = img_it(im)
                          except(ValueError):
                              uvs = np.abs(im.uv)
                              bms,dim,dbm = uvs,uvs,uvs
                          for k in ['uvs','bms','dim','dbm']:
                              if k in outputs: to_fits(k, eval(k), s,imgcnt,history=history)
                          imgcnt += 1
                      us,vs,ws,ds,wgts = [],[],[],[],[]
                      if opts.no_w:
                          im = a.img.Img(opts.size, opts.res, mf_order=0)
                      else:
                          im = a.img.ImgW(opts.size, opts.res, mf_order=0, wres=opts.wres)
                      if opts.src == 'zen':
                          s = a.phs.RadioFixedBody(aa.sidereal_time(), 
                              aa.lat, name='zen')
                          src = a.fit.SrcCatalog([s])
              curtime = t
              aa.set_jultime(t)
              src.compute(aa)
              if s.alt < opts.altmin * a.img.deg2rad: continue
              s_eq = src.get_crds('eq', ncrd=3)
              aa.sim_cache(s_eq)
          if s.alt < opts.altmin * a.img.deg2rad: continue
          aa.set_active_pol(pol)
          d,f = d.take(chans), f.take(chans)
          if not opts.skip_amp: d /= aa.passband(i,j)
          # Throws PointingError if not up:
          if not opts.skip_phs:
              if opts.zen_phased: d = aa.unphs2src(d, 'z', i, j)
              d = aa.phs2src(d, s, i, j)
          u,v,w = aa.gen_uvw(i,j,src=s)
          longenough = np.where(np.sqrt(u**2+v**2) < opts.minuv, 0, 1).squeeze()
          if not opts.skip_bm:
              # Calculate beam strength for weighting purposes
              wgt = aa.bm_response(i,j).squeeze()
              # Optimal SNR: down-weight beam-attenuated data 
              # by another factor of the beam response.
              d *= wgt; wgt *= wgt
          else: wgt = np.ones(d.shape, dtype=np.float)
          valid = np.logical_and(np.logical_not(f), longenough)
          d = d.compress(valid)
          if len(d) == 0: continue
          n_ints += 1
          ds.append(d)
          us.append(u.compress(valid))
          vs.append(v.compress(valid))
          ws.append(w.compress(valid))
          wgts.append(wgt.compress(valid))
          # If data buffer is full, grid data
          if len(ds) * len(chans) > opts.buf_thresh:
              grid_it(im,us,vs,ws,ds,wgts)
              us,vs,ws,ds,wgts = [],[],[],[],[]

    # Grid remaining data into UV matrix
    try:
        grid_it(im,us,vs,ws,ds,wgts)
        uvs,bms,dim,dbm = img_it(im)
    except(ValueError):
        print 'No data: skipping output file.'
        continue
    for k in ['uvs','bms','dim','dbm']:
        if k in outputs: to_fits(k, eval(k), s, imgcnt,history=history)
    imgcnt += 1
    us,vs,ws,ds,wgts = [],[],[],[],[]
    if opts.no_w:
        im = a.img.Img(opts.size, opts.res, mf_order=0)
    else:
        im = a.img.ImgW(opts.size, opts.res, mf_order=0, wres=opts.wres)
    




#! /usr/bin/env python
"""
This is a general-purpose script for making faceted, spherical maps (stored in
Healpix FITS format) from individual "flat" maps stored in FITS files.

Author: Aaron Parsons
"""

import sys, numpy as np, os, aipy as a, optparse, ephem

o = optparse.OptionParser()
o.set_usage('mk_map.py [options] *.fits')
o.set_description(__doc__)
o.add_option('-i', '--interpolate', dest='interpolate', action='store_true',
    help='Use sub-pixel interpolation when gridding data to healpix map.')
o.add_option('-m', '--map', dest='map',
    help='The skymap file to use.  If it exists, new data will be added to the map.  Othewise, the file will be created.')
o.add_option('--nside', dest='nside', type='int', default=256,
    help='NSIDE parameter for map, if creating a new file.')
o.add_option('--fwidth', dest='fwidth', type='float', default=10,
    help='Width of gaussian, in degrees, used to downweight data away from pointing center for each facet.  Default 10.')
opts, args = o.parse_args(sys.argv[1:])

# Open skymap
if os.path.exists(opts.map): skymap = a.map.Map(fromfits=opts.map)
else: skymap = a.map.Map(nside=opts.nside)
skymap.set_interpol(opts.interpolate)

prev_dra = None
for i, filename in enumerate(args):
    img, kwds = a.img.from_fits(filename)
    img = img.squeeze()
    # Read ra/dec of image center, which are stored in J2000
    assert(kwds['epoch'] == 2000)
    s = ephem.Equatorial(kwds['ra']*a.img.deg2rad, kwds['dec']*a.img.deg2rad, 
        epoch=ephem.J2000)
    # To precess the entire image to J2000, we actually need to precess the
    # coords of the center back to the epoch of the observation (obs_date),
    # get the pixel coordinates in that epoch, and then precess the coords of
    # each pixel.  This is because extrapolating pixel coords from the J2000 
    # center assumes the J2000 ra/dec axes, which may be tilted relative to
    # the ra/dec axes of the epoch of the image.
    s = ephem.Equatorial(s, epoch=kwds['obs_date'])
    ra, dec = s.get()
    print '-----------------------------------------------------------'
    print 'Reading file %s (%d / %d)' % (filename, i + 1, len(args))
    print kwds
    print 'Pointing (ra, dec):', ra, dec
    print 'Image Power:', np.abs(img).sum()
    if prev_dra != kwds['d_ra']:
        prev_dra = kwds['d_ra']
        DIM = img.shape[0]
        RES = 1. / (kwds['d_ra'] * a.img.deg2rad * DIM)
        im = a.img.Img(DIM*RES, RES, mf_order=0)
        tx,ty,tz = im.get_top(center=(DIM/2,DIM/2))
        # Define a weighting for gridding data into the skymap
        map_wgts = np.exp(-(tx**2+ty**2) / np.sin(opts.fwidth*a.img.deg2rad)**2)
        map_wgts.shape = (map_wgts.size,)
        valid = np.logical_not(map_wgts.mask)
        map_wgts = map_wgts.compress(valid)
    # Get coordinates of image pixels in the epoch of the observation
    ex,ey,ez = im.get_eq(ra, dec, center=(DIM/2,DIM/2))
    ex = ex.compress(valid); ey = ey.compress(valid); ez = ez.compress(valid)
    img = img.flatten(); img = img.compress(valid)
    # Precess the pixel coordinates to the (J2000) epoch of the map
    m = a.coord.convert_m('eq','eq', 
        iepoch=kwds['obs_date'], oepoch=ephem.J2000)
    ex,ey,ez = np.dot(m, np.array([ex,ey,ez])) 
    # Put the data into the skymap
    skymap.add((ex,ey,ez), map_wgts, img)
skymap.to_fits(opts.map, clobber=True)




#! /usr/bin/env python
"""
A script for reducing the number of channels in a UV data set by coherently
adding adjacent channels together.

Author: Aaron Parsons
Date: 6/03/07
Revisions:
    12/11/07    arp Ported to new miriad file interface
"""

import aipy as a, sys, os, numpy as np, optparse

o = optparse.OptionParser()
o.set_usage('combine_freqs.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-n', '--nchan', dest='nchan', default=256, type='int',
    help='Reduce the number of channels in a spectrum to this number.')
o.add_option('-c', '--careful_flag', dest='careful_flag', action='store_true',
    help='Flag resultant bin if any component bins are flagged (otherwise, flags only when there are more flagged bins than unflagged bins).')
o.add_option('-d', '--dont_flag', dest='dont_flag', action='store_true',
    help='Only flag resultant bin if every component bin is flagged (otherwise, uses any data available).')
o.add_option('-u', '--unify', dest='unify', action='store_true',
    help='Output to a single UV file.')
opts, args = o.parse_args(sys.argv[1:])

uvo = None
for uvfile in args:
    print uvfile,'->',uvfile+'m'
    uvi = a.miriad.UV(uvfile)
    sfreq,sdf,nchan = uvi['sfreq'], uvi['sdf'], uvi['nchan']
    newsfreq = sfreq + ((nchan/opts.nchan) * sdf) / 2
    newsdf = sdf * (nchan/opts.nchan)
    
    if uvo is None:
        uvofile = uvfile+'m'
        if os.path.exists(uvofile):
            print uvofile, 'exists, skipping.'
            continue
        uvo = a.miriad.UV(uvofile, status='new')
        if nchan != opts.nchan:
            uvo.init_from_uv(uvi, override={'nchan':opts.nchan, 
                'sfreq':newsfreq, 'sdf':newsdf, 'nschan':opts.nchan, 
                'freq':newsfreq, 'nchan0':opts.nchan, },)
        else: uvo.init_from_uv(uvi)

    if nchan != opts.nchan:
        def mfunc(uv, p, d, f):
            d = np.where(f, 0, d)
            d.shape = (opts.nchan, nchan/opts.nchan)
            d = d.sum(axis=1)
            f.shape = (opts.nchan, nchan/opts.nchan)
            f = np.logical_not(f).astype(np.int).sum(axis=1)
            d /= f.clip(1,np.Inf)
            if opts.careful_flag: f = np.where(f < nchan/opts.nchan, 1, 0)
            elif opts.dont_flag: f = np.where(f < 1, 1, 0)
            else: f = np.where(f <= nchan/opts.nchan/2, 1, 0)
            return p, np.where(f, 0, d), f
        uvo.pipe(uvi, mfunc=mfunc, raw=True,
            append2hist='COMB_FREQ: nchan=%d careful=%s dont=%s unify=%s\n' % \
                (opts.nchan, opts.careful_flag, opts.dont_flag, opts.unify))
    else:
        uvo.pipe(uvi, 
            append2hist='COMB_FREQ: nchan=%d careful=%s dont=%s unify=%s\n' % \
                (opts.nchan, opts.careful_flag, opts.dont_flag, opts.unify))
    if not opts.unify:
        del(uvo)
        uvo = None
    



"""
This package collects together tools for radio astronomical interferometry.
In addition to pure-python phasing, calibration, imaging, and
deconvolution code, this package includes interfaces to MIRIAD (a Fortran
interferometry package) and HEALPix (a package for representing spherical
data sets), and some math/fitting routines from SciPy.
All code provided is released under the GNU General Public License
(see LICENSE.txt).

Author: Aaron Parsons
"""

import phs, const, coord, deconv
import ephem, fit, healpix, img
import interp, cal, map, miriad
import optimize, rfi, amp, scripting, src, _src, utils
import dsp
import pol, twodgauss #added by dfm
from __gitlog__ import __gitlog__
from __version__ import __version__
from __branch__ import __branch__



# -*- coding: utf-8 -*-

"""Unit test suite for the AIPY package."""

import unittest

import _alm_test
import _healpix_test
import amp_test
import coord_test
import deconv_test
import miriad_test
import phs_test
import phs_benchmark
import scripting_test

class TestSuite(unittest.TestSuite):
        """A unittest.TestSuite class which contains all of the package unit tests."""

        def __init__(self):
                """Setup the AIPY package unit test suite."""

                unittest.TestSuite.__init__(self)

                self.addTest(_alm_test.TestSuite())
                self.addTest(_healpix_test.TestSuite())
                self.addTest(amp_test.TestSuite())
                self.addTest(coord_test.TestSuite())
                self.addTest(deconv_test.TestSuite())
                self.addTest(miriad_test.TestSuite())
                self.addTest(phs_test.TestSuite())
                self.addTest(phs_benchmark.TestSuite())
                self.addTest(scripting_test.TestSuite())

def main(opts=None, args=None):
    """Function to call all of the lsl tests."""

    if opts is not None:
        if opts.verbose:
            level = 2
        else:
            level = 1
    else:
        level = 2

    suite = TestSuite()
    runner = unittest.TextTestRunner(verbosity = level)
    runner.run(suite)


if __name__  == '__main__':
    import optparse

    parser = optparse.OptionParser(usage = "python %prog [options]", description = __doc__)
    parser.add_option("-v", "--verbose", action = "store_true", dest = "verbose", default = False,
                      help = "extra print output")
    (opts, args) = parser.parse_args()

    main(opts, args)



'''This module interfaces to the Helmboldt catalog (http://arxiv.org/abs/0707.3418)'''

import aipy as a, numpy as np, os

class HelmboldtFixedBody(a.fit.RadioFixedBody):
    def compute(self, observer):
        a.phs.RadioFixedBody.compute(self, observer)
        self.update_jys(observer.get_afreqs())
    def update_jys(self, afreqs):
        A = np.log10(self._jys)
        try: B,C,D = (list(self.index) + [0,0,0])[:3]
        except(TypeError): B,C,D = (self.index,0,0)
        X = np.log10(afreqs / self.mfreq)
        self.jys = 10**(A + B*X + C*np.exp(D*X))
    def get_params(self, prm_list=None):
        """Return all fitable parameters in a dictionary."""
        aprms = {
            'jys':      float(self._jys),
            'ra':       float(self._ra),
            'dec':      float(self._dec),
            'a1':       float(self.srcshape[0]),
            'a2':       float(self.srcshape[1]),
            'th':       float(self.srcshape[2]),
            'dra':      float(self.ionref[0]),
            'ddec':     float(self.ionref[1]),
        }
        try: aprms['index'] = list(self.index)
        except(TypeError): aprms['index'] = float(self.index)
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms

class HelmboldtCatalog(a.fit.SrcCatalog):
    metadata = {}
    def get_metadata(self):
        '''Return info from posfile about constituent measurements from which
        spectral fits were obtained.  Returns dictionary with source names
        linked to lists of (freq,flux,error) triplets. Must first run get_catalog
        to retrieve data.'''
        return self.metadata
    rms = {}
    def get_rms(self):
        """Return a dictionary giving the Helmboldt fit rms for each source.
        rms is only computed where enought data below 325MHz is available to 
        compute a spectral index or Kuehr fit. Null value is indicated with
        a rms=-1. Must first run get_catalog to retrieve data."""
        return self.rms
    ncomp = {}
    def get_ncomp(self):
        """Return a dictionary giving the Helmboldt fit rms for each source.
        rms is only computed where enought data below 325MHz is available to 
        compute a spectral index or Kuehr fit. Null value is indicated with
        a rms=-1. Must first run get_catalog to retrieve data."""
        return self.ncomp
    def fromfile(self, posfile, fitfile):
        srcs = {}
        rms = {}
        ncomp = {}
        # Read RA/DEC
        srclines = [L for L in open(posfile).readlines() if L.startswith('J')]
        for line in srclines:
            srcname = line[:9]
            srcs[srcname] = line[35:57]
            if not self.ncomp.has_key(srcname): self.ncomp[srcname] = int(line[33])
            if not self.metadata.has_key(srcname): self.metadata[srcname] = []
            md = (float(line[58:64])/1e3,float(line[65:73]),float(line[74:81]))
            self.metadata[srcname].append(md)
        for s in srcs:
            ra = srcs[s][:10].strip().replace(' ',':')
            dec = srcs[s][11:].strip().replace(' ',':')
            srcs[s] = [ra, dec]
        # Read spectral data
        srclines = [L for L in open(fitfile).readlines() if L.startswith('J')]
        for line in srclines: 
            srcs[line[:9]].append(map(float, line[13:62].split()))
            try: self.rms[line[:9]] = float(line[63:70])
            except(ValueError): self.rms[line[:9]]=-1

        addsrcs = []
        for s in srcs:
            ra,dec,spec = srcs[s]
            # If there's not a good fit on data, use VLSS value and default index
            if len(spec) < 5 or spec[3] == -99.:
                srctype = a.fit.RadioFixedBody
                jys = spec[0]
                # If there is no index put bogus value of -99
                try: index = spec[1]
                except(IndexError): index = -99
            else:
                srctype = HelmboldtFixedBody
                ABCD = spec[3:]
                jys,index = 10**ABCD[0], ABCD[1:]
            addsrcs.append(srctype(ra, dec, name=s, 
                jys=jys, index=index, mfreq=.074))
        self.add_srcs(addsrcs)

FITFILE = os.path.dirname(__file__) + os.sep + 'helm_fit.txt'
POSFILE = os.path.dirname(__file__) + os.sep + 'helm_pos.txt'

_helmcat = None

def get_srcs(srcs=None, cutoff=None):
    # Mechanism for delaying instantiation of catalog until it is accessed
    global _helmcat
    if _helmcat is None:
        _helmcat = HelmboldtCatalog()
        _helmcat.fromfile(POSFILE, FITFILE)
    if srcs is None:
        if cutoff is None: srcs = _helmcat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _helmcat.keys(): _helmcat[s].update_jys(fq)
            srcs = [s for s in _helmcat.keys() if _helmcat[s].jys[0] > cut]
    srclist = []
    for s in srcs:
        try: srclist.append(_helmcat[s])
        except(KeyError): pass
    return srclist



'''The Mauritius Radio Telescope catalog.  Can be downloaded from
http://www.rri.res.in/surveys/MRT/Catalogues/CatalogueMRTa.txt Copy this file
to "CatalogueMRTa.txt" in the _src directory of your AIPY installation.'''

import aipy as a, numpy as np, os

class MRTCatalog(a.fit.SrcCatalog):
    def fromfile(self, filename):
        f = open(filename)
        addsrcs = []
        for L in f.readlines()[4:]:
            s, ra, dec, jys = L.split()[:4]
            addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=s,
                jys=float(jys), index=0, mfreq=.150))
        self.add_srcs(addsrcs)

MRTFILE = os.path.dirname(__file__) + os.sep + 'CatalogueMRTa.txt'
_mrtcat = None

def get_srcs(srcs=None, cutoff=None):
    # Mechanism for delaying instantiation of catalog until it is accessed
    global _mrtcat
    if _mrtcat is None:
        _mrtcat = MRTCatalog()
        _mrtcat.fromfile(MRTFILE)
    if srcs is None:
        if cutoff is None: srcs = _mrtcat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _mrtcat.keys(): _mrtcat[s].update_jys(fq)
            srcs = [s for s in _mrtcat.keys() if _mrtcat[s].jys[0] > cut]
    srclist = []
    for s in srcs:
        try: srclist.append(_mrtcat[s])
        except(KeyError): pass
    return srclist




#!/usr/bin/python
"""
Parse large data files by antenna number. Keep antenna 0 in all of them as a phase reference.
"""

import aipy as a
import sys,os,optparse

o = optparse.OptionParser()
o.set_usage('pull_ants.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-a','--ants',dest='ants',type='str',default=None,help='List of Antennae to pull.')
opts,args = o.parse_args(sys.argv[1:])

if opts.ants == None:
    print 'You forgot to tell me which antennae to pull. Try using a -a ant1,ant2,ant3 option or something, dude'
    sys.exit()

for filename in args:
    outfile = filename+'A'
    print filename,'-->',outfile
    if os.path.exists(outfile):
        print 'File exists, skipping'
        continue
    
    ants2use = opts.ants.split(',') 
    
    def mfunc(uv,p,d):
        uvw,t,(i,j) = p
        if i in ants2use and j in ants2use: return p,d
        else: return None,None
    
    uvi = a.miriad.UV(filename)
    uvo = a.miriad.UV(outfile,status='new')
    uvo.init_from_uv(uvi)
    histstr = 'PULL ANTS: ants='%opts.refant
    for ant in ants2use:
        histstr += str(ant)+','
        if ant == ants2use[-1]: histstr += '\n'
    uvo.pipe(uvi,mfunc=mfunc,append2hist=histstr)
    del uvo,uvi



#!/usr/bin/env python

"""
Do one iteration of a delay-solving algorithm.
"""

import aipy as a
import numpy as np
from matplotlib import pylab as pl
import optparse,sys,os
from time import time
from glob import glob

o = optparse.OptionParser()
o.set_description(__doc__)
o.set_usage('find_delay.py [options] *.uv.src')
o.add_option('-C','--cal',dest='cal',default=None,help='Cal file to use')
o.add_option('-f','--flag',dest='flag',type='str',help='Antennae to flag. Ex. 1,2,3')
o.add_option('--fixant',dest='fixant',type='str',default=None,help='Antennae whose delays you want to fix at calfile value')
o.add_option('--refant',dest='refant',default=0,help='Reference Antenna Number (default=0)')
o.add_option('-p','--plots',dest='plots',default=None,action='store_true',help='Plots?')
opts,args = o.parse_args(sys.argv[1:])

t0 = time()

#Parse Arguments
FlagAnt = []
if opts.flag:
    for antstr in opts.flag.split(','):
        FlagAnt.append(int(antstr))

exec('from %s import prms' % opts.cal)
        

#Read in Data
DD = {}
AntNos = []
for uvfile in args:
    print 'Reading',uvfile
    uv = a.miriad.UV(uvfile)
    for (uvw,t,(i,j)),d in uv.all():
        if i == j: continue
        pol = a.miriad.pol2str[uv['pol']]
        if pol == 'xy' : continue
        if pol == 'yx' : continue
        if i in FlagAnt: continue
        if j in FlagAnt: continue
        if not pol in DD.keys(): DD[pol] = {}
        if not (i,j) in DD[pol].keys(): DD[pol][(i,j)] = np.zeros_like(d)
        #Construct flagging kernel
        flags = np.logical_not(d.mask).astype(np.float)
        gain = np.sqrt(np.average(flags**2))
        ker = np.fft.ifft(flags)
        d = d.filled(0)
        #Delay transform
        d = np.fft.ifft(d)
        d = np.ma.array(d)
        d = np.ma.concatenate([d[d.shape[0]/2:],d[:d.shape[0]/2]],axis=0)
        DD[pol][(i,j)] += np.abs(d)
        if not i in AntNos: AntNos.append(i)
        if not j in AntNos: AntNos.append(j)
delays = np.arange(-0.5/uv['sdf'],0.5/uv['sdf'],1/(uv['sdf']*uv['nchan']))
pols = DD.keys()
AntNos.sort()
pols.sort()
del(uv)

#Pick out maximum values DO A BETTER JOB OF THIS!!!
Phi_ij = {}
for pol in pols:
    if not pol in Phi_ij.keys(): Phi_ij[pol] = {}
    for ij in DD[pol].keys():
        for i,x in enumerate(DD[pol][ij]):
            if x == np.max(DD[pol][ij]): Phi_ij[pol][ij] = delays[i]
del(DD)

#Construct matrices to do linear algebra and shit.
Npol = len(Phi_ij.keys())
Nant = len(AntNos)
Nbl = int(Nant*(Nant-1)/2)


#Set delays you want to fix.
if opts.refant == None: refant = AntNos[0]
else: refant = opts.refant

if not opts.fixdelay: FixDelay = {}
if opts.fixdelay == 'all':
    FixDelay = {}
    for pol in pols:
        try: FixDelay[pol] = [prms['delays'][i][pol] for i in AntNos]
        except(KeyError): FixDelay[pol] = [prms['delays'][i] for i in AntNos]
        FixDelayIndex[pi] = range(len(AntNos))
else: 
    FixDelay = {}
    FixDelayIndex = []
    for pol in pols:
        FixDelay[pol] = []
        for i,ant in AntNos:
            if ant in opts.fixdelay.split(','):
                try: FixDelay[pol].append(prms['delays'][ant][pol])
                except(KeyError): FixDelay[pol].append(prms['delays'][ant])
                if not i in FixDelayIndex: FixDelayIndex.append(i)
            
Tau = {}
CC = {}
for pol in pols:
    
    A,b = {},{}
    A = np.zeros((Nbl+len(FixDelay[pol])+1,Nant))
    b = np.zeros(Nbl+len(FixDelay[pol])+1)

    #populatin the countryside, populatin the villages!
    index = 0
    for i,ant1 in enumerate(AntNos):
        if ant1 == refant:
            A[-1,i] = 1.
            b[-1] = 0.
        for j,ant2 in enumerate(AntNos):
            if j <= i: continue
            b[index] = Phi_ij[pol][(ant1,ant2)]
            A[index,i],A[index,j] = 1.,-1.    
            index += 1
    #Set Constraints
    for i,Di in enumerate(FixDelayIndex):
        A[Nbl+i,Di] = 1.
        b[Nbl+i] = FixDelay[pol][i]
    Tau[pol] = np.dot(np.linalg.inv(np.dot(A.T,A)),np.dot(A.T,b))
    #These are the delays (in ns)!!!

    #Compute closure phases
    CC[pol] = {}
    for i in AntNos:
        for j in AntNos:
            if j <= i: continue
            for k in AntNos:
                if k <= j: continue
                CC[pol][(i,j,k)] = Phi_ij[pol][(i,j)] + Phi_ij[pol][(j,k)] - Phi_ij[pol][(i,k)]

print 'Computing time =', time() - t0,'s for',Nbl,'baselines.'

#Print delays:
for pol in pols:
    for i,d in enumerate(Tau[pol]):
        badstr = ''
        if np.abs(d-np.average(Tau[pol])) >= 2.*np.std(Tau[pol]): badstr = '!!!'
        print 'Antenna',AntNos[i],pol[0],':',d,'ns',badstr


#Generate plots:
if opts.plots:
    #Delay vs. Antenna
    for pol in pols:
        pl.plot(AntNos,Tau[pol],'o',label=pol)
        pl.xlabel('Antenna Number')
        pl.ylabel('Delay (ns)')
    pl.legend()
    pl.show()

    #Histogram of all Closures
    c = {}
    for pol in CC:
        if not pol in c: c[pol] = {}
        for (i,j,k) in CC[pol]:
            if not 'all' in c[pol]: c[pol]['all'] = []
            if not i in c[pol]: c[pol][i] = []
            if not j in c[pol]: c[pol][j] = []
            if not k in c[pol]: c[pol][k] = []
            c[pol]['all'].append(CC[pol][(i,j,k)])
            c[pol][i].append(CC[pol][(i,j,k)])
            c[pol][j].append(CC[pol][(i,j,k)])
            c[pol][k].append(CC[pol][(i,j,k)])

    for i,pol in enumerate(c.keys()):
        for ant in c[pol]:
            pl.hist(c[pol][ant],bins=50)
            pl.title('Antenna %s, %s' % (ant,pol))
            pl.xlabel('Closure phase (ns)')
            pl.ylabel('Number')
            pl.show()
            



import unittest
import aipy._dsp as _dsp
import numpy as n

class Testgrid1D_c(unittest.TestCase):
    def test_sanity(self):
        buf = n.zeros(32, dtype=n.complex64)
        ind = n.array([5, 10.1, 14.9], dtype=n.float32)
        dat = n.array([1, 1, 1], dtype=n.complex64)
        _dsp.grid1D_c(buf, ind, dat)
        x = n.arange(32)
        ans = n.exp(-(x-5)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2) + \
            n.exp(-(x-10.1)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2) + \
            n.exp(-(x-14.9)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2)
        if False:
            from matplotlib import pylab as P
            P.plot(ind, n.abs(dat), '.')
            P.plot(n.abs(buf))
            P.semilogy(x, n.exp(-(x-5)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            P.semilogy(x, n.exp(-(x-10.1)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            P.semilogy(x, n.exp(-(x-14.9)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            P.semilogy(x, n.abs(ans))
            P.ylim(1e-10, 1)
            P.show()
        self.assertAlmostEqual(n.max(n.abs(buf - ans)), 0, 6)

class Testgrid2D_c(unittest.TestCase):
    def test_sanity(self):
        buf = n.zeros((32,32), dtype=n.complex64)
        ind = n.array([[5,5], [10.1,10.1], [14.5, 15.5]], dtype=n.float32)
        ind1 = ind[:,0].copy()  # copy necessary b/c not contigous memory otherwise
        ind2 = ind[:,1].copy()
        dat = n.array([1, 1, 1], dtype=n.complex64)
        _dsp.grid2D_c(buf, ind1, ind2, dat)
        self.assertAlmostEqual(buf[5,5], 0.63661977236758149, 6)
        self.assertAlmostEqual(buf[10,10], 0.63661977236758149 * n.exp(-2*(.1**2+.1**2)), 6)
        self.assertAlmostEqual(buf[14,15], 0.63661977236758149 * n.exp(-2*(.5**2+.5**2)), 6)
        if False:
            from matplotlib import pylab as P
            P.imshow(n.log10(n.abs(buf)), vmax=0, vmin=-6, interpolation='nearest')
            #P.plot(n.abs(buf))
            #P.semilogy(x, n.exp(-(x-5)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            #P.semilogy(x, n.exp(-(x-10.1)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            #P.semilogy(x, n.exp(-(x-14.9)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            #P.semilogy(x, n.abs(ans))
            #P.ylim(1e-10, 1)
            P.show()

class Testdegrid2D_c(unittest.TestCase):
    def test_sanity(self):
        buf = n.ones((32,32), dtype=n.complex64)
        ind = n.array([[5,5], [10.1,10.1], [14.5, 15.5]], dtype=n.float32)
        ind1 = ind[:,0].copy()  # copy necessary b/c not contigous memory otherwise
        ind2 = ind[:,1].copy()
        dat = n.zeros(ind1.shape, dtype=n.complex64)
        _dsp.degrid2D_c(buf, ind1, ind2, dat)
        self.assertTrue(n.all(dat == 1))
        if False:
            from matplotlib import pylab as P
            P.imshow(n.log10(n.abs(buf)), vmax=0, vmin=-6, interpolation='nearest')
            #P.plot(n.abs(buf))
            #P.semilogy(x, n.exp(-(x-5)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            #P.semilogy(x, n.exp(-(x-10.1)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            #P.semilogy(x, n.exp(-(x-14.9)**2 / (2*.5**2))/n.sqrt(2*n.pi*.5**2))
            #P.semilogy(x, n.abs(ans))
            #P.ylim(1e-10, 1)
            P.show()

if __name__ == '__main__':
    unittest.main()



"""
An astronomy library for precessing coordinates between epochs and converting 
between topocentric (z = up, x = east), ecliptic (heliocentric), equatorial
(celestial), and galactic coordinate systems.  Vectors are 3 dimensional 
with magnitude 1, representing a point on the unit sphere.  Includes generic 
3 vector rotation code.
"""

import numpy as np
import ephem as e

sys_dict = {
    'eq': e.Equatorial,
    'ec': e.Ecliptic,
    'ga': e.Galactic
}

def convert(crd, isys, osys, iepoch=e.J2000, oepoch=e.J2000):
    """Convert 'crd' from coordinate system isys to osys, including
    epoch precession.  Valid coordinate systems are 'ec' (Ecliptic), 'eq' 
    (Equatorial), and 'ga' (Galactic).  Epochs may be date strings, or
    numerical ephemeris times."""
    if len(crd) == 3: crd = eq2radec(crd)
    c1 = sys_dict[isys[:2].lower()](crd[0], crd[1], epoch=iepoch)
    return sys_dict[osys[:2].lower()](c1, epoch=oepoch).get()

def convert_m(isys, osys, iepoch=e.J2000, oepoch=e.J2000):
    """Return the 3x3 matrix corresponding to a coordinate 
    system/precssion transformation (see 'convert'). 
    NOTE: To obtain a transformation matrix to dot with isys to get osys, 
    must reverse isys and osys arguments in convert_m. """
    m = np.array([[1,0,0],[0,1,0],[0,0,1]], dtype=np.double)
    for i in range(3):
        c = convert(m[:,i], isys, osys, iepoch=iepoch, oepoch=oepoch)
        m[:,i] = radec2eq(c)
    return m

def rot_m(ang, vec):
    """Return 3x3 matrix defined by rotation by 'ang' around the
    axis 'vec', according to the right-hand rule.  Both can be vectors,
    returning a vector of rotation matrices.  Rotation matrix will have a 
    scaling of |vec| (i.e. normalize |vec|=1 for a pure rotation)."""
    c = np.cos(ang); s = np.sin(ang); C = 1-c
    x,y,z = vec[...,0], vec[...,1], vec[...,2]
    xs,ys,zs = x*s, y*s, z*s
    xC,yC,zC = x*C, y*C, z*C
    xyC,yzC,zxC = x*yC, y*zC, z*xC
    rm = np.array([[x*xC+c, xyC-zs, zxC+ys],
                   [xyC+zs, y*yC+c, yzC-xs],
                   [zxC-ys, yzC+xs, z*zC+c]], dtype=np.double)
    if rm.ndim > 2:
        axes = range(rm.ndim)
        return rm.transpose(axes[-1:] + axes[:-1])
    else:
        return rm

def xyz2thphi(xyz):
    """Convert xyz vectors (x,y,z along first axis) into angles theta
    (from z axis), phi (counter-clockwise around z, 0 at x axis)."""
    x,y,z = xyz
    phi = np.arctan2(y, x)
    th = np.arctan2(np.sqrt(x**2+y**2),z)
    if np.ma.isMA(x):
        try:
            return np.ma.array([th.filled(0),phi.filled(0)], 
                mask=[x.mask,x.mask], dtype=np.double)
        except(np.core.ma.MAError):
            return np.ma.array([th,phi], dtype=np.double)
    return np.array([th,phi], dtype=np.double)

def thphi2xyz(th_phi):
    """Convert angles theta (from z axis), phi (counter-clockwise around z, 
    0 at x axis) into xyz vectors (x,y,z along first axis)."""
    th,phi = th_phi
    z = np.cos(th)
    r = np.sin(th)
    x,y = r*np.cos(phi), r*np.sin(phi)
    if np.ma.isMA(th):
        try:
            return np.ma.array([x.filled(),y.filled(),z.filled()], 
                mask=[th.mask, th.mask, th.mask], dtype=np.double)
        except(np.core.ma.MAError):
            return np.ma.array([x,y,z], dtype=np.double)
    return np.array([x,y,z], dtype=np.double)

def eq2radec(xyz):
    """Convert equatorial xyz vectors (x,y,z along first axis) into angles ra 
    (counter-clockwise around z = north, 0 at x axis), dec (from equator)."""
    th_phi = xyz2thphi(xyz)
    th,phi = th_phi
    dec = np.pi/2 - th
    ra = np.ma.where(phi < 0, phi + 2*np.pi, phi)
    th_phi[0],th_phi[1] = ra,dec
    return th_phi

def radec2eq(ra_dec):
    """Convert angles ra (counter-clockwise around z = north, 0 at x axis), dec 
    (from equator) into equatorial xyz vectors (x,y,z along first axis)."""
    phi,th = ra_dec
    return thphi2xyz((np.pi/2 - th, phi))

def latlong2xyz(lat_long):
    """Convert angles lat (from equator), long (counter-clockwise around
    z = north, 0 at x axis) into xyz vectors (x,y,z along first axis)."""
    lat,long = lat_long
    return radec2eq((long,lat))

def top2azalt(xyz):
    """Convert topocentric xyz vectors (x,y,z along first axis) into angles az 
    (clockwise around z = up, 0 at x axis = north), alt (from horizon)."""
    th_phi = xyz2thphi(xyz)
    th,phi = th_phi
    alt = np.pi/2 - th
    az = np.pi/2 - phi
    az = np.ma.where(az < 0, az + 2*np.pi, az)
    th_phi[0],th_phi[1] = az,alt
    return th_phi

def azalt2top(az_alt):
    """Convert angles az (clockwise around z = up, 0 at x axis = north), alt 
    (from horizon) into topocentric xyz vectors (x,y,z along first axis)."""
    az,alt = az_alt
    return thphi2xyz((np.pi/2 - alt, np.pi/2 - az))

def eq2top_m(ha, dec):
    """Return the 3x3 matrix converting equatorial coordinates to topocentric
    at the given hour angle (ha) and declination (dec)."""
    sin_H, cos_H = np.sin(ha), np.cos(ha)
    sin_d, cos_d = np.sin(dec), np.cos(dec)
    zero = np.zeros_like(ha)
    map =  np.array([[    sin_H    ,       cos_H  ,       zero  ],
                     [ -sin_d*cos_H,   sin_d*sin_H,      cos_d  ],
                     [  cos_d*cos_H,  -cos_d*sin_H,      sin_d  ]])
    if len(map.shape) == 3: map = map.transpose([2, 0, 1])
    return map

def top2eq_m(ha, dec):
    """Return the 3x3 matrix converting topocentric coordinates to equatorial
    at the given hour angle (ha) and declination (dec)."""
    m = eq2top_m(ha, dec)
    if len(m.shape) == 3:
        for i in range(m.shape[0]): m[i] = np.linalg.inv(m[i])
        return m
    else: return np.linalg.inv(eq2top_m(ha, dec))




#! /usr/bin/env python
"""
This is a general-purpose script for plotting simple FITS images.
"""

import aipy as a, sys, optparse, os
import numpy as np, ephem, math
from matplotlib import pylab as p

o = optparse.OptionParser()
o.set_usage('plot_img.py [options] *.fits')
o.set_description(__doc__)
a.scripting.add_standard_options(o, chan=True, cmap=True, max=True, drng=True)
o.add_option('-m', '--mode', dest='mode', default='log',
    help='Plot mode can be log (logrithmic), lin (linear), phs (phase), real, or imag.')
o.add_option('-o', '--outfile', dest='outfile', default='',
    help='If provided, will save the figure to the specified file instead of popping up a window.')
o.add_option('-p', '--pol', dest='pol', type='int', default=0, 
    help='Polarization index if FITS file has multiple polarizations.  Default 0.')
o.add_option('--batch', dest='batch', action='store_true',
    help='Process files in batch mode (one plot each) and output to a <input file>.png file')
o.add_option('--nogrid', dest='nogrid', action='store_true',
    help='Do not display RA/DEC grid.')
o.add_option('-f', '--fft', dest='fft', action='store_true',
    help='Perform 2D FFT of image.')
opts, args = o.parse_args(sys.argv[1:])

cmap = p.get_cmap(opts.cmap)
if opts.batch: m1,m2 = 1,1
else:
    m2 = int(math.sqrt(len(args)))
    m1 = int(math.ceil(float(len(args)) / m2))

for cnt, filename in enumerate(args):
    print filename
    if opts.batch:
        cnt = 0
        outfile = filename+'.png'
        if os.path.exists(outfile):
            print 'Output file exists... skipping.'
            continue
    # Gather data
    d, kwds = a.img.from_fits(filename)
    print d.shape
    print kwds
    print '-----------------------------------------------------------'

    # Parse command-line options
    compress_axes = []
    ra_ax,dec_ax = (0,1)
    try:
        for i,ax in enumerate(kwds['axes']):
            if ax.startswith('ra'): ra_ax = i
            elif ax.startswith('dec'): dec_ax = i
            elif ax.startswith('freq'):
                chans = a.scripting.parse_chans(opts.chan, d.shape[i])
                d = d.take(chans, axis=i)
                compress_axes.append(i)
            elif ax.startswith('stokes'):
                d = d.take([opts.pol], axis=i)
                compress_axes.append(i)
    except(KeyError): pass

    compress_axes.reverse()
    for ax in compress_axes: d = np.average(d, axis=ax)

    # Put array in (ra,dec) order for plotting
    d = d.transpose((ra_ax,dec_ax))

    # Generate plots
    if opts.fft:
        d = np.fft.fft2(d)
        d = a.img.recenter(d, (d.shape[0]/2, d.shape[1]/2))
    if opts.mode.startswith('phs'): d = np.angle(d)
    elif opts.mode.startswith('lin'): d = np.absolute(d)
    elif opts.mode.startswith('real'): d = d.real
    elif opts.mode.startswith('imag'): d = d.imag
    elif opts.mode.startswith('log'):
        d = np.ma.absolute(d)
        d = np.ma.masked_less_equal(d, 0)
        d = np.ma.log10(d)

    if not opts.max is None: max = opts.max
    else: max = d.max()
    if not opts.drng is None: min = max - opts.drng
    else: min = d.min()

    p.subplot(m2, m1, cnt+1)
    if not opts.nogrid:
        from mpl_toolkits.basemap import Basemap
        xpx,ypx = d.shape
        dx1 = -(xpx/2 + .5) * kwds['d_ra'] * a.img.deg2rad
        dx2 = (xpx/2 - .5) * kwds['d_ra'] * a.img.deg2rad
        dy1 = -(ypx/2 + .5) * kwds['d_dec'] * a.img.deg2rad
        dy2 = (ypx/2 - .5) * kwds['d_dec'] * a.img.deg2rad
        map = Basemap(projection='ortho', lon_0=180, lat_0=kwds['dec'],
            rsphere=1, llcrnrx=dx1, llcrnry=dy1, urcrnrx=dx2,urcrnry=dy2)
        map.drawmeridians(np.arange(kwds['ra']-180,kwds['ra']+180,30))
        map.drawparallels(np.arange(-90,120,30))
        map.drawmapboundary()
        map.imshow(d, vmin=min, vmax=max, cmap=cmap, interpolation='nearest')
    else: p.imshow(d, vmin=min, vmax=max, origin='lower', cmap=cmap, interpolation='nearest')
    p.colorbar(shrink=.5, fraction=.05)
    p.title(filename)

    if opts.batch:
        print 'Saving to', outfile
        p.savefig(outfile)
        p.clf()
        

# Add right-click functionality for finding locations/strengths in map.
cnt = 1
def click(event):
    global cnt
    if not event.button in [2,3]: return
    if not opts.nogrid:
        lon,lat = map(event.xdata, event.ydata, inverse=True)
        lon = (180 + kwds['ra'] - lon) % 360
        lon *= a.img.deg2rad; lat *= a.img.deg2rad
        ra,dec = ephem.hours(lon), ephem.degrees(lat)
        xpx = np.around((event.xdata-1-dx1) / (dx2 - dx1) * d.shape[0] - .5)
        ypx = np.around((event.ydata-1-dy1) / (dy2 - dy1) * d.shape[1] - .5)
        flx = d[ypx,xpx]
        if opts.mode.startswith('log'): flx = 10**flx
        print '#%d (RA,DEC): (%s, %s), PX: (%d,%d) Jy: %f' % (cnt, ra, dec, xpx, ypx, flx)
    else:
        xpx = np.around(event.xdata)
        ypx = np.around(event.ydata)
        flx = d[ypx,xpx]
        if opts.mode.startswith('log'): flx = 10**flx
        print '#%d PX: (%d,%d) Jy: %f' % (cnt, xpx, ypx, flx)
    cnt += 1

#register this function with the event handler
p.connect('button_press_event', click)

if not opts.batch:
    if opts.outfile != '':
        print 'Saving to', opts.outfile
        p.savefig(opts.outfile)
    else: p.show()



'''The WENSS Catalog. 
Data files are in tab-separated format from Vizier.
To download in the correct format, open a catalog online in Vizier,
select'Tab-Separated-Values' as the Output layout in the drop-down box, set
the maximum entries to 'unlimited', and click 'Sexagesimal' under the box
for 'Target Name or Position'.  Submit the query, and copy the output to a
txt file.  Copy this file to "wenss.txt" in the _src directory of your AIPY
installation.'''
import aipy as a, numpy as np, os

class WenssCatalog(a.fit.SrcCatalog):
    def fromfile(self,filename):
        f = open(filename)
        addsrcs = []
        for L in [L for L in f.readlines() if not L.startswith('#')]:
            text = L.split('\t')
            if len(text) <= 4: continue
            try: int(text[0][0])
            except(ValueError): continue
            ra = text[4].replace(' ',':')
            dec = text[5].replace(' ',':')
            name = text[2].strip()
            jys = float(text[9])/1000
            addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                jys=jys, index=0, mfreq=0.330))
        self.add_srcs(addsrcs)

WENSSFILE = os.path.dirname(__file__) + os.sep + 'wenss.txt'
_wensscat = None

def get_srcs(srcs=None, cutoff=None):
    global _wensscat
    if _wensscat is None:
        _wensscat = WenssCatalog()
        _wensscat.fromfile(WENSSFILE)
    if srcs is None:
        if cutoff is None: srcs = _wensscat.keys()
        else:
            cut, fq = cutoff
            fq = np.array([fq])
            for s in _wensscat.keys(): _wensscat[s].update_jys(fq)
            srcs = [s for s in _wensscat.keys() if _wensscat[s].jys[0] > cut]
    srclist = []
    for s in srcs:
        try: srclist.append(_wensscat[s])
        except(KeyError): pass
    return srclist
